"""Infer and retrieve whether a node in our AST denotes a Golang pointer."""
from typing import Optional, Final, MutableMapping, List, Mapping, Union

from aas_core_codegen import intermediate
from aas_core_codegen.common import (
    Error,
    assert_never,
)
from aas_core_codegen.intermediate import (
    type_inference as intermediate_type_inference,
)
from aas_core_codegen.parse import tree as parse_tree

_OPTIONAL_PRIMITIVE_INFERRED_TYPE_IS_POINTER = {
    intermediate_type_inference.PrimitiveType.BOOL: True,
    intermediate_type_inference.PrimitiveType.INT: True,
    intermediate_type_inference.PrimitiveType.FLOAT: True,
    intermediate_type_inference.PrimitiveType.STR: True,
    intermediate_type_inference.PrimitiveType.BYTEARRAY: False,
    intermediate_type_inference.PrimitiveType.LENGTH: True,
    intermediate_type_inference.PrimitiveType.NONE: False,
}

assert all(
    a_type in _OPTIONAL_PRIMITIVE_INFERRED_TYPE_IS_POINTER
    for a_type in intermediate_type_inference.PrimitiveType
)

_OPTIONAL_PRIMITIVE_TYPE_IS_POINTER = {
    intermediate.PrimitiveType.BOOL: True,
    intermediate.PrimitiveType.INT: True,
    intermediate.PrimitiveType.FLOAT: True,
    intermediate.PrimitiveType.STR: True,
    intermediate.PrimitiveType.BYTEARRAY: False,
}

assert all(
    a_type in _OPTIONAL_PRIMITIVE_TYPE_IS_POINTER
    for a_type in intermediate.PrimitiveType
)


def is_pointer_type(
    type_annotation: Union[
        intermediate_type_inference.TypeAnnotationUnion,
        intermediate.TypeAnnotationUnion,
    ]
) -> bool:
    """Check whether ``type_annotation`` denotes a pointer type."""
    if isinstance(type_annotation, intermediate_type_inference.TypeAnnotation):
        if not isinstance(
            type_annotation, intermediate_type_inference.OptionalTypeAnnotation
        ):
            return False

        if isinstance(
            type_annotation.value, intermediate_type_inference.PrimitiveTypeAnnotation
        ):
            return _OPTIONAL_PRIMITIVE_INFERRED_TYPE_IS_POINTER[
                type_annotation.value.a_type
            ]

        elif isinstance(
            type_annotation.value, intermediate_type_inference.OurTypeAnnotation
        ):
            if isinstance(type_annotation.value.our_type, intermediate.Enumeration):
                return True

            elif isinstance(
                type_annotation.value.our_type, intermediate.ConstrainedPrimitive
            ):
                return _OPTIONAL_PRIMITIVE_INFERRED_TYPE_IS_POINTER[
                    intermediate_type_inference.PRIMITIVE_TYPE_MAP[
                        type_annotation.value.our_type.constrainee
                    ]
                ]
            else:
                return False
        else:
            return False

    elif isinstance(type_annotation, intermediate.TypeAnnotation):
        # NOTE (mristin, 2023-05-24):
        # This is almost identical as the previous if-branch, but not completely.
        # Note the map ``_OPTIONAL_PRIMITIVE_TYPE_IS_POINTER``.

        if not isinstance(type_annotation, intermediate.OptionalTypeAnnotation):
            return False

        if isinstance(type_annotation.value, intermediate.PrimitiveTypeAnnotation):
            return _OPTIONAL_PRIMITIVE_TYPE_IS_POINTER[type_annotation.value.a_type]

        elif isinstance(type_annotation.value, intermediate.OurTypeAnnotation):
            if isinstance(type_annotation.value.our_type, intermediate.Enumeration):
                return True

            elif isinstance(
                type_annotation.value.our_type, intermediate.ConstrainedPrimitive
            ):
                return _OPTIONAL_PRIMITIVE_TYPE_IS_POINTER[
                    type_annotation.value.our_type.constrainee
                ]
            else:
                return False
        else:
            return False

    else:
        assert_never(type_annotation)


class Inferrer(parse_tree.Transformer[Optional[Error]]):
    """
    Infer the types of the given parse tree.

    Since we also handle non-nullness, you need to pre-compute the canonical
    representation of the nodes that you want to infer the types for. To that end,
    use :class:`~CanonicalRepresenter`.
    """

    #: Track of the inferred is-pointer features; ``True`` if the type of the node is
    #: represented as a pointer type in Golang.
    is_pointer_map: Final[MutableMapping[parse_tree.Node, bool]]

    #: Errors encountered during the inference
    errors: Final[List[Error]]

    def __init__(
        self,
        environment: intermediate_type_inference.Environment,
        type_map: Mapping[
            parse_tree.Node, intermediate_type_inference.TypeAnnotationUnion
        ],
    ) -> None:
        """Initialize with the given values."""
        # We need to create our own child environment so that we can introduce new
        # entries without affecting the variables from the outer scopes.
        self._environment = intermediate_type_inference.MutableEnvironment(
            parent=environment
        )

        self._type_map = type_map

        self.is_pointer_map = dict()
        self.errors = []

    def transform_member(self, node: parse_tree.Member) -> Optional[Error]:
        error = self.transform(node.instance)
        if error is not None:
            return error

        instance_type_anno = intermediate_type_inference.beneath_optional(
            self._type_map[node.instance]
        )

        if isinstance(
            instance_type_anno,
            intermediate_type_inference.EnumerationAsTypeTypeAnnotation,
        ):
            self.is_pointer_map[node] = False
            return None

        if not (
            isinstance(
                instance_type_anno, intermediate_type_inference.OurTypeAnnotation
            )
        ):
            error = Error(
                node.instance.original_node,
                f"Unexpected {instance_type_anno}; expected the instance of "
                f"a member access to be annotated with our type",
            )
            self.errors.append(error)
            return error

        our_type = instance_type_anno.our_type

        if not isinstance(
            our_type, (intermediate.AbstractClass, intermediate.ConcreteClass)
        ):
            error = Error(
                node.instance.original_node,
                f"Unexpected {instance_type_anno.our_type}; expected the instance of "
                f"a member access to be annotated as an abstract or a concrete class",
            )
            self.errors.append(error)
            return error

        prop = our_type.properties_by_name.get(node.name, None)
        if prop is not None:
            self.is_pointer_map[node] = is_pointer_type(prop.type_annotation)
            return None

        method = our_type.methods_by_name.get(node.name, None)
        if method is not None:
            self.is_pointer_map[node] = False
            return None

        error = Error(
            node.original_node,
            f"The member {node.name} not found in the class {our_type.name!r}",
        )
        self.errors.append(error)
        return error

    def transform_index(self, node: parse_tree.Index) -> Optional[Error]:
        error = self.transform(node.collection)
        if error is not None:
            return error

        error = self.transform(node.index)
        if error is not None:
            return error

        collection_type_anno = intermediate_type_inference.beneath_optional(
            self._type_map[node.collection]
        )

        if not isinstance(
            collection_type_anno, intermediate_type_inference.ListTypeAnnotation
        ):
            error = Error(
                node.collection.original_node,
                f"Expected the collection to be a list in the index, "
                f"but got: {collection_type_anno}",
            )
            self.errors.append(error)
            return error

        items_type_anno = intermediate_type_inference.beneath_optional(
            collection_type_anno.items
        )

        self.is_pointer_map[node] = is_pointer_type(items_type_anno)
        return None

    def transform_comparison(self, node: parse_tree.Comparison) -> Optional[Error]:
        last_error = None  # type: Optional[Error]
        for operand in (node.left, node.right):
            # NOTE (mristin, 2023-05-24):
            # Do not immediately return so that other arguments are processed as well.
            # This way we get a longer list of errors which the caller can report
            # using :py:prop:`errors`.

            error = self.transform(operand)
            if error is not None:
                last_error = error

        if last_error is not None:
            return last_error

        self.is_pointer_map[node] = False
        return None

    def transform_is_in(self, node: parse_tree.IsIn) -> Optional[Error]:
        error = self.transform(node.member)
        if error is not None:
            return error

        error = self.transform(node.container)
        if error is not None:
            return error

        self.is_pointer_map[node] = False
        return None

    def transform_implication(self, node: parse_tree.Implication) -> Optional[Error]:
        error = self.transform(node.antecedent)
        if error is not None:
            return error

        error = self.transform(node.consequent)
        if error is not None:
            return error

        self.is_pointer_map[node] = False
        return None

    def transform_method_call(self, node: parse_tree.MethodCall) -> Optional[Error]:
        last_error = None  # type: Optional[Error]
        for arg in node.args:
            error = self.transform(arg)

            # NOTE (mristin, 2023-05-24):
            # Do not immediately return so that other arguments are processed as well.
            # This way we get a longer list of errors which the caller can report
            # using :py:prop:`errors`.
            if error is not None:
                last_error = error

        error = self.transform(node.member)
        if error is not None:
            last_error = error

        if last_error is not None:
            return last_error

        instance_type_anno = intermediate_type_inference.beneath_optional(
            self._type_map[node.member.instance]
        )

        if not (
            isinstance(
                instance_type_anno, intermediate_type_inference.OurTypeAnnotation
            )
        ):
            error = Error(
                node.member.instance.original_node,
                f"Unexpected {instance_type_anno}; expected the instance of "
                f"a member access to be annotated with our type",
            )
            self.errors.append(error)
            return error

        our_type = instance_type_anno.our_type

        if not isinstance(
            our_type, (intermediate.AbstractClass, intermediate.ConcreteClass)
        ):
            error = Error(
                node.member.instance.original_node,
                f"Unexpected {instance_type_anno.our_type}; expected the instance of "
                f"a member access to be annotated as an abstract or a concrete class",
            )
            self.errors.append(error)
            return error

        method = our_type.methods_by_name[node.member.name]
        if method.returns is None:
            self.is_pointer_map[node] = False
        else:
            self.is_pointer_map[node] = is_pointer_type(method.returns)

        return None

    def transform_function_call(self, node: parse_tree.FunctionCall) -> Optional[Error]:
        last_error = None  # type: Optional[Error]
        for arg in node.args:
            error = self.transform(arg)

            # NOTE (mristin, 2023-05-24):
            # Do not immediately return so that other arguments are processed as well.
            # This way we get a longer list of errors which the caller can report
            # using :py:prop:`errors`.
            if error is not None:
                last_error = error

        error = self.transform(node.name)
        if error is not None:
            last_error = error

        if last_error is not None:
            return last_error

        func_type = self._type_map[node.name]

        if not isinstance(
            func_type,
            (
                intermediate_type_inference.VerificationTypeAnnotation,
                intermediate_type_inference.BuiltinFunctionTypeAnnotation,
            ),
        ):
            error = Error(
                node.name.original_node,
                f"Expected the function to be either "
                f"{intermediate_type_inference.VerificationTypeAnnotation.__name__} or "
                f"{intermediate_type_inference.BuiltinFunctionTypeAnnotation}, "
                f"but got {func_type}",
            )
            self.errors.append(error)
            return error

        if func_type.func.returns is None:
            self.is_pointer_map[node] = False
        else:
            self.is_pointer_map[node] = is_pointer_type(func_type.func.returns)

        return None

    def transform_constant(self, node: parse_tree.Constant) -> Optional[Error]:
        self.is_pointer_map[node] = False
        return None

    def transform_is_none(self, node: parse_tree.IsNone) -> Optional[Error]:
        error = self.transform(node.value)
        if error is not None:
            return error

        self.is_pointer_map[node] = False
        return None

    def transform_is_not_none(self, node: parse_tree.IsNotNone) -> Optional[Error]:
        error = self.transform(node.value)
        if error is not None:
            return error

        self.is_pointer_map[node] = False
        return None

    def transform_not(self, node: parse_tree.Not) -> Optional[Error]:
        error = self.transform(node.operand)
        if error is not None:
            return error

        self.is_pointer_map[node] = False
        return None

    def transform_name(self, node: parse_tree.Name) -> Optional[Error]:
        type_in_env = self._environment.find(node.identifier)
        if type_in_env is None:
            error = Error(
                node.original_node,
                f"We do not know how to infer the type of "
                f"the variable with the identifier {node.identifier!r} from the "
                f"given environment. Mind that we do not consider the module "
                f"scope nor handle all built-in functions due to simplicity! If "
                f"you believe this needs to work, please notify the developers.",
            )
            self.errors.append(error)
            return error

        self.is_pointer_map[node] = is_pointer_type(type_in_env)
        return None

    def transform_and(self, node: parse_tree.And) -> Optional[Error]:
        last_error = None  # type: Optional[Error]
        for value in node.values:
            # NOTE (mristin, 2023-05-24):
            # Do not immediately return so that other arguments are processed as well.
            # This way we get a longer list of errors which the caller can report
            # using :py:prop:`errors`.

            error = self.transform(value)
            if error is not None:
                last_error = error

        if last_error is not None:
            return last_error

        self.is_pointer_map[node] = False
        return None

    def transform_or(self, node: parse_tree.Or) -> Optional[Error]:
        last_error = None  # type: Optional[Error]
        for value in node.values:
            # NOTE (mristin, 2023-05-24):
            # Do not immediately return so that other arguments are processed as well.
            # This way we get a longer list of errors which the caller can report
            # using :py:prop:`errors`.

            error = self.transform(value)
            if error is not None:
                last_error = error

        if last_error is not None:
            return last_error

        self.is_pointer_map[node] = False
        return None

    def transform_add(self, node: parse_tree.Add) -> Optional[Error]:
        last_error = None  # type: Optional[Error]
        for operand in (node.left, node.right):
            # NOTE (mristin, 2023-05-24):
            # Do not immediately return so that other arguments are processed as well.
            # This way we get a longer list of errors which the caller can report
            # using :py:prop:`errors`.

            error = self.transform(operand)
            if error is not None:
                last_error = error

        if last_error is not None:
            return last_error

        self.is_pointer_map[node] = False
        return None

    def transform_sub(self, node: parse_tree.Sub) -> Optional[Error]:
        last_error = None  # type: Optional[Error]
        for operand in (node.left, node.right):
            # NOTE (mristin, 2023-05-24):
            # Do not immediately return so that other arguments are processed as well.
            # This way we get a longer list of errors which the caller can report
            # using :py:prop:`errors`.

            error = self.transform(operand)
            if error is not None:
                last_error = error

        if last_error is not None:
            return last_error

        self.is_pointer_map[node] = False
        return None

    def transform_formatted_value(
        self, node: parse_tree.FormattedValue
    ) -> Optional[Error]:
        error = self.transform(node.value)
        if error is not None:
            return error

        self.is_pointer_map[node] = False
        return None

    def transform_joined_str(self, node: parse_tree.JoinedStr) -> Optional[Error]:
        last_error = None  # type: Optional[Error]
        for value in node.values:
            if isinstance(value, str):
                continue

            # NOTE (mristin, 2023-05-24):
            # Do not immediately return so that other arguments are processed as well.
            # This way we get a longer list of errors which the caller can report
            # using :py:prop:`errors`.
            error = self.transform(value)
            if error is not None:
                last_error = error

        if last_error is not None:
            return last_error

        self.is_pointer_map[node] = False
        return None

    def transform_for_each(self, node: parse_tree.ForEach) -> Optional[Error]:
        variable_type_in_env = self._environment.find(node.variable.identifier)
        if variable_type_in_env is not None:
            error = Error(
                node.variable.original_node,
                f"The variable {node.variable.identifier} "
                f"has been already defined before",
            )
            self.errors.append(error)
            return error

        loop_variable_type = self._type_map[node.variable]
        self.is_pointer_map[node.variable] = is_pointer_type(loop_variable_type)

        self.is_pointer_map[node] = False
        return None

    def transform_for_range(self, node: parse_tree.ForRange) -> Optional[Error]:
        # noinspection PyUnusedLocal
        error = None  # type: Optional[Error]

        variable_type_in_env = self._environment.find(node.variable.identifier)
        if variable_type_in_env is not None:
            error = Error(
                node.variable.original_node,
                f"The variable {node.variable.identifier} "
                f"has been already defined before",
            )
            self.errors.append(error)
            return error

        last_error = None  # type: Optional[Error]
        for value in (node.start, node.end):
            error = self.transform(value)

            # NOTE (mristin, 2023-05-24):
            # Do not immediately return so that other arguments are processed as well.
            # This way we get a longer list of errors which the caller can report
            # using :py:prop:`errors`.

            if error is not None:
                last_error = error

        if last_error is not None:
            return last_error

        loop_variable_type = self._type_map[node.variable]
        self.is_pointer_map[node.variable] = is_pointer_type(loop_variable_type)

        self.is_pointer_map[node] = False
        return None

    def _transform_any_or_all(
        self, node: Union[parse_tree.Any, parse_tree.All]
    ) -> Optional[Error]:
        error = self.transform(node.generator)
        if error is not None:
            return error

        try:
            loop_variable_type = self._type_map[node.generator.variable]

            self._environment.set(
                identifier=node.generator.variable.identifier,
                type_annotation=loop_variable_type,
            )

            error = self.transform(node.condition)
            if error is not None:
                return error
        finally:
            self._environment.remove(identifier=node.generator.variable.identifier)

        self.is_pointer_map[node] = False
        return None

    def transform_any(self, node: parse_tree.Any) -> Optional[Error]:
        return self._transform_any_or_all(node)

    def transform_all(self, node: parse_tree.All) -> Optional[Error]:
        return self._transform_any_or_all(node)

    def transform_assignment(self, node: parse_tree.Assignment) -> Optional[Error]:
        error = self.transform(node.value)
        if error is not None:
            return error

        error = self.transform(node.target)
        if error is not None:
            return error

        self.is_pointer_map[node] = False
        return None

    def transform_return(self, node: parse_tree.Return) -> Optional[Error]:
        # Just recurse to fill ``type_map`` on ``value`` even though we know the type
        # in advance
        if node.value is not None:
            error = self.transform(node.value)
            if error is not None:
                return error

        self.is_pointer_map[node] = False
        return None



"""Provide common functions shared among different Python code generation modules."""
import enum
import io
import re
from typing import List, cast, Tuple, Optional, Mapping

from icontract import ensure, require

from aas_core_codegen import intermediate
from aas_core_codegen.common import Stripped, assert_never, Identifier
from aas_core_codegen.python import naming as python_naming


class StringQuoting(enum.Enum):
    """Represent how strings should be quoted as Python literals."""

    SINGLE_QUOTES = 0
    DOUBLE_QUOTES = 1


# region Different string escapings

# See: https://python-reference.readthedocs.io/en/latest/docs/str/escapes.html
_BASE_ESCAPING_IN_PYTHON = {
    "\\": "\\\\",
    "\a": "\\a",
    "\b": "\\b",
    "\f": "\\f",
    "\n": "\\n",
    "\r": "\\r",
    "\t": "\\t",
    "\v": "\\v",
}

_ESCAPING_IN_PYTHON_INCLUDING_DOUBLE_QUOTES = {
    **_BASE_ESCAPING_IN_PYTHON,
    **{'"': '\\"'},
}

_ESCAPING_IN_PYTHON_INCLUDING_DOUBLE_QUOTES_AND_DUPLICATE_CURLY_BRACKETS = {
    **_ESCAPING_IN_PYTHON_INCLUDING_DOUBLE_QUOTES,
    **{
        "{": "{{",
        "}": "}}",
    },
}

_ESCAPING_IN_PYTHON_INCLUDING_SINGLE_QUOTES = {
    **_BASE_ESCAPING_IN_PYTHON,
    **{"'": "\\'"},
}

_ESCAPING_IN_PYTHON_INCLUDING_SINGLE_QUOTES_AND_DUPLICATE_CURLY_BRACKETS = {
    **_ESCAPING_IN_PYTHON_INCLUDING_SINGLE_QUOTES,
    **{
        "{": "{{",
        "}": "}}",
    },
}

# endregion


# fmt: off
@ensure(
    lambda quoting, without_enclosing, result:
    not (quoting is StringQuoting.SINGLE_QUOTES and not without_enclosing)
    or (
        result.startswith("'") and result.endswith("'")
    )
)
@ensure(
    lambda quoting, without_enclosing, result:
    not (quoting is StringQuoting.DOUBLE_QUOTES and not without_enclosing)
    or (
        result.startswith('"') and result.endswith('"')
    )
)
# fmt: on
def string_literal(
    text: str,
    quoting: Optional[StringQuoting] = None,
    without_enclosing: bool = False,
    duplicate_curly_brackets: bool = False,
) -> Stripped:
    """
    Generate a string literal from the ``text``.

    If ``quoting`` is not set, check which quotes occur more often (single-quotes or
    double-quotes), and enclose the literal such that we need to escape as little as
    possible.

    If ``without_enclosing`` is set, the enclosing characters (double-quotes or
    single-quotes) are omitted.

    If ``duplicate_curly_brackets`` is set, all the opening and closing curly brackets
    (``{`` and ``}``) are duplicated (``{{`` and ``}}``, respectively).
    """
    mapping: Mapping[str, str]
    enclosing: str

    if quoting is None:
        if text.count("'") <= text.count('"'):
            mapping = _ESCAPING_IN_PYTHON_INCLUDING_SINGLE_QUOTES
            enclosing = "'"

        else:
            mapping = _ESCAPING_IN_PYTHON_INCLUDING_DOUBLE_QUOTES
            enclosing = '"'

    elif quoting is StringQuoting.SINGLE_QUOTES:
        mapping = _ESCAPING_IN_PYTHON_INCLUDING_SINGLE_QUOTES
        enclosing = "'"

    elif quoting is StringQuoting.DOUBLE_QUOTES:
        mapping = _ESCAPING_IN_PYTHON_INCLUDING_DOUBLE_QUOTES
        enclosing = '"'
    else:
        assert_never(quoting)

    if duplicate_curly_brackets:
        if mapping is _ESCAPING_IN_PYTHON_INCLUDING_DOUBLE_QUOTES:
            mapping = (
                _ESCAPING_IN_PYTHON_INCLUDING_DOUBLE_QUOTES_AND_DUPLICATE_CURLY_BRACKETS
            )
        elif mapping is _ESCAPING_IN_PYTHON_INCLUDING_SINGLE_QUOTES:
            mapping = (
                _ESCAPING_IN_PYTHON_INCLUDING_SINGLE_QUOTES_AND_DUPLICATE_CURLY_BRACKETS
            )
        else:
            raise AssertionError(f"Unexpected mapping: {mapping}")

    escaped = "".join(mapping.get(character, character) for character in text)

    if without_enclosing:
        return Stripped(escaped)
    else:
        return Stripped(f"{enclosing}{escaped}{enclosing}")


# fmt: off
@ensure(
    lambda result:
    not result[1] or all(
        line.startswith('b"') and line.endswith('"')
        for line in result[0].splitlines()
    ),
    "If multi-line, the text is lines of bytes literals"
)
@ensure(
    lambda result:
    not (not result[1])
    or (
        '\n' not in result[0]
        and result[0].startswith('b"') and result[0].endswith('"')
    )
)
# fmt: on
def bytes_literal(value: bytes) -> Tuple[Stripped, bool]:
    """
    Generate a literal representing the ``value``.

    If there are more than 8 bytes, a multi-line literal is returned.

    :param value: to be represented
    :return: (Python literal, is multi-line)
    """
    writer = io.StringIO()

    multi_line: bool

    if len(value) <= 8:
        writer.write('b"')
        for byte in value:
            writer.write(f"\\x{byte:02x}")

        writer.write('"')

        multi_line = False
    else:
        for i in range(0, len(value), 8):
            if i > 0:
                writer.write("\n")

            start = i
            end = min(i + 8, len(value))

            assert start < end

            writer.write('b"')
            for byte in value[start:end]:
                writer.write(f"\\x{byte:02x}")
            writer.write('"')

        multi_line = True

    return Stripped(writer.getvalue()), multi_line


def needs_escaping(text: str, also_check_curly_brackets: bool = False) -> bool:
    """
    Check whether the ``text`` contains a character that needs escaping.

    If ``also_check_curly_brackets`` is set, it also checks that the ``text``
    does not contain any curly brackets, which would need to be properly escaped
    in string interpolations.
    """
    for character in text:
        if character == "\a":
            return True
        elif character == "\b":
            return True
        elif character == "\f":
            return True
        elif character == "\n":
            return True
        elif character == "\r":
            return True
        elif character == "\t":
            return True
        elif character == "\v":
            return True
        elif character == '"':
            return True
        elif character == "\\":
            return True
        else:
            pass

    if also_check_curly_brackets:
        if "{" in text:
            return True

        if "}" in text:
            return True

    return False


PRIMITIVE_TYPE_MAP = {
    intermediate.PrimitiveType.BOOL: Stripped("bool"),
    intermediate.PrimitiveType.INT: Stripped("int"),
    intermediate.PrimitiveType.FLOAT: Stripped("float"),
    intermediate.PrimitiveType.STR: Stripped("str"),
    # NOTE (mristin, 2022-10-14):
    # Since most Python functions and encodings deal with ``bytes`` instead of
    # ``bytearrays``, we decided to use ``bytes`` in the SDK.
    intermediate.PrimitiveType.BYTEARRAY: Stripped("bytes"),
}


def _assert_all_primitive_types_are_mapped() -> None:
    """Assert that all the primitive types are mapped to Python ones."""
    all_primitive_literals = set(literal.value for literal in PRIMITIVE_TYPE_MAP)

    mapped_primitive_literals = set(
        literal.value for literal in intermediate.PrimitiveType
    )

    all_diff = all_primitive_literals.difference(mapped_primitive_literals)
    mapped_diff = mapped_primitive_literals.difference(all_primitive_literals)

    messages = []  # type: List[str]
    if len(mapped_diff) > 0:
        messages.append(
            f"More primitive maps are mapped than there were defined "
            f"in the ``intermediate._types``: {sorted(mapped_diff)}"
        )

    if len(all_diff) > 0:
        messages.append(
            f"One or more primitive types in the ``intermediate._types`` were not "
            f"mapped in PRIMITIVE_TYPE_MAP: {sorted(all_diff)}"
        )

    if len(messages) > 0:
        raise AssertionError("\n\n".join(messages))


_assert_all_primitive_types_are_mapped()


def generate_type(
    type_annotation: intermediate.TypeAnnotationUnion,
    types_module: Optional[Identifier] = None,
) -> Stripped:
    """
    Generate the type for the given type annotation.

    If ``types_module`` is specified, it is used as prefix for the composite types.
    """
    if isinstance(type_annotation, intermediate.PrimitiveTypeAnnotation):
        return PRIMITIVE_TYPE_MAP[type_annotation.a_type]

    elif isinstance(type_annotation, intermediate.OurTypeAnnotation):
        our_type = type_annotation.our_type

        # NOTE (mristin, 2022-09-17):
        # If no ``types_module``, we mark all enumerations and classes as string
        # literals to avoid problems caused by lack of forward declaration in Python. If
        # we created a dependency graph, we could strip away some quotes, but we believe
        # that consistency is better for readability than no quotes in some cases.

        if isinstance(our_type, intermediate.Enumeration):
            if types_module is None:
                return Stripped(
                    repr(python_naming.enum_name(type_annotation.our_type.name))
                )
            else:
                return Stripped(
                    f"{types_module}"
                    f".{python_naming.enum_name(type_annotation.our_type.name)}"
                )

        elif isinstance(our_type, intermediate.ConstrainedPrimitive):
            return PRIMITIVE_TYPE_MAP[our_type.constrainee]

        elif isinstance(our_type, intermediate.Class):
            if types_module is None:
                return Stripped(repr(python_naming.class_name(our_type.name)))
            else:
                return Stripped(
                    f"{types_module}.{python_naming.class_name(our_type.name)}"
                )

    elif isinstance(type_annotation, intermediate.ListTypeAnnotation):
        item_type = generate_type(
            type_annotation=type_annotation.items, types_module=types_module
        )

        return Stripped(f"List[{item_type}]")

    elif isinstance(type_annotation, intermediate.OptionalTypeAnnotation):
        value = generate_type(
            type_annotation=type_annotation.value, types_module=types_module
        )

        return Stripped(f"Optional[{value}]")

    else:
        assert_never(type_annotation)

    raise AssertionError("Should not have gotten here")


INDENT = "    "
INDENT2 = INDENT * 2
INDENT3 = INDENT * 3
INDENT4 = INDENT * 4
INDENT5 = INDENT * 5
INDENT6 = INDENT * 6

WARNING = Stripped(
    """\
# This code has been automatically generated by aas-core-codegen.
# Do NOT edit or append."""
)

QUALIFIED_MODULE_NAME_RE = re.compile(
    r"[a-zA-Z_][a-zA-Z_0-9]*(\.[a-zA-Z_][a-zA-Z_0-9]*)*"
)


class QualifiedModuleName(str):
    """Capture a qualified name of a module."""

    @require(lambda identifier: QUALIFIED_MODULE_NAME_RE.fullmatch(identifier))
    def __new__(cls, identifier: str) -> "QualifiedModuleName":
        return cast(QualifiedModuleName, identifier)


class GeneratorForLoopVariables:
    """
    Generate a unique variable name based on ``item`` stem.

    >>> generator = GeneratorForLoopVariables()

    >>> next(generator)
    'an_item'

    >>> next(generator)
    'another_item'

    >>> next(generator)
    'yet_another_item'

    >>> next(generator)
    'yet_yet_another_item'
    """

    def __init__(self) -> None:
        """Initialize with the zero counter."""
        self.counter = 0

    def __next__(self) -> Identifier:
        """Generate the next variable name."""
        if self.counter == 0:
            result = Identifier("an_item")
        elif self.counter == 1:
            result = Identifier("another_item")
        else:
            result = Identifier(("yet_" * (self.counter - 1)) + "another_item")

        self.counter += 1

        return result



"""Emulate naming logic for our types to Protocol Buffers in Python code."""

from aas_core_codegen.common import Identifier
from aas_core_codegen import naming


def enum_name(identifier: Identifier) -> Identifier:
    """
    Generate a name for an enum based on its meta-model ``identifier``.

    >>> enum_name(Identifier("something"))
    'Something'

    >>> enum_name(Identifier("URL_to_something"))
    'UrlToSomething'
    """
    return naming.capitalized_camel_case(identifier)


def enum_literal_constant_name(
    enumeration_name: Identifier, literal_name: Identifier
) -> Identifier:
    """
    Generate a constant identifier for an enum literal in the meta-model.

    >>> enum_literal_constant_name(
    ...     Identifier("that"), Identifier("literal")
    ... )
    'That_LITERAL'

    >>> enum_literal_constant_name(
    ...     Identifier("URL_to_something"), Identifier("ABC_shines")
    ... )
    'Urltosomething_ABC_SHINES'

    >>> enum_literal_constant_name(
    ...     Identifier("Modelling_kind"), Identifier("Template")
    ... )
    'Modellingkind_TEMPLATE'
    """
    enum_name_joined = "".join(enumeration_name.split("_"))

    return Identifier(f"{enum_name_joined.capitalize()}_{literal_name.upper()}")


def property_name(identifier: Identifier) -> Identifier:
    """
    Generate the name of the property in a Protocol Buffer generated class.

    >>> property_name(Identifier("something"))
    'something'

    >>> property_name(Identifier("something_to_URL"))
    'something_to_url'

    >>> property_name(Identifier("URL_to_something"))
    'url_to_something'
    """
    return naming.lower_snake_case(identifier)


def class_name(identifier: Identifier) -> Identifier:
    """
    Generate a name for a class based on its meta-model ``identifier``.

    >>> class_name(Identifier("something"))
    'Something'

    >>> class_name(Identifier("URL_to_something"))
    'UrlToSomething'
    """
    return naming.capitalized_camel_case(identifier)


def choice_class_name(identifier: Identifier) -> Identifier:
    """
    Generate a name for a choice (union) class based on its meta-model ``identifier``.

    >>> choice_class_name(Identifier("something"))
    'Something_choice'

    >>> choice_class_name(Identifier("URL_to_something"))
    'UrlToSomething_choice'
    """
    return Identifier(naming.capitalized_camel_case(identifier) + "_choice")



"""Provide the intermediate representation of the meta-model."""

from aas_core_codegen.intermediate import _types, _translate, _stringify

TypeAnnotation = _types.TypeAnnotation
TypeAnnotationUnion = _types.TypeAnnotationUnion
AtomicTypeAnnotation = _types.AtomicTypeAnnotation
PrimitiveType = _types.PrimitiveType
PRIMITIVE_TYPE_TO_PYTHON_TYPE = _types.PRIMITIVE_TYPE_TO_PYTHON_TYPE
PYTHON_TYPE_TO_PRIMITIVE_TYPE = _types.PYTHON_TYPE_TO_PRIMITIVE_TYPE
PrimitiveTypeAnnotation = _types.PrimitiveTypeAnnotation
OurTypeAnnotation = _types.OurTypeAnnotation
ListTypeAnnotation = _types.ListTypeAnnotation
OptionalTypeAnnotation = _types.OptionalTypeAnnotation
SummaryRemarksDescription = _types.SummaryRemarksDescription
SummaryRemarksConstraintsDescription = _types.SummaryRemarksConstraintsDescription
DescriptionOfMetaModel = _types.DescriptionOfMetaModel
DescriptionOfOurType = _types.DescriptionOfOurType
DescriptionOfProperty = _types.DescriptionOfProperty
DescriptionOfEnumerationLiteral = _types.DescriptionOfEnumerationLiteral
DescriptionOfSignature = _types.DescriptionOfSignature
DescriptionOfConstant = _types.DescriptionOfConstant
DescriptionUnion = _types.DescriptionUnion
Property = _types.Property
Default = _types.Default
DefaultPrimitive = _types.DefaultPrimitive
DefaultEnumerationLiteral = _types.DefaultEnumerationLiteral
Argument = _types.Argument
OurType = _types.OurType
OurTypeExceptEnumeration = _types.OurTypeExceptEnumeration
Invariant = _types.Invariant
Contract = _types.Contract
Snapshot = _types.Snapshot
Contracts = _types.Contracts
Method = _types.Method
ImplementationSpecificMethod = _types.ImplementationSpecificMethod
UnderstoodMethod = _types.UnderstoodMethod
MethodUnion = _types.MethodUnion
Constructor = _types.Constructor
Serialization = _types.Serialization
EnumerationLiteral = _types.EnumerationLiteral
Enumeration = _types.Enumeration
ConstrainedPrimitive = _types.ConstrainedPrimitive
Class = _types.Class
ClassUnion = _types.ClassUnion
ConcreteClass = _types.ConcreteClass
AbstractClass = _types.AbstractClass
ConstantPrimitive = _types.ConstantPrimitive
PrimitiveSetLiteral = _types.PrimitiveSetLiteral
Constant = _types.Constant
ConstantSetOfPrimitives = _types.ConstantSetOfPrimitives
ConstantSetOfEnumerationLiterals = _types.ConstantSetOfEnumerationLiterals
ConstantSetUnion = _types.ConstantSetUnion
ConstantUnion = _types.ConstantUnion
Verification = _types.Verification
ImplementationSpecificVerification = _types.ImplementationSpecificVerification
PatternVerification = _types.PatternVerification
TranspilableVerification = _types.TranspilableVerification
VerificationUnion = _types.VerificationUnion
Signature = _types.Signature
Interface = _types.Interface
SymbolTable = _types.SymbolTable

type_annotations_equal = _types.type_annotations_equal
beneath_optional = _types.beneath_optional
TypeAnnotationExceptOptional = _types.TypeAnnotationExceptOptional
try_primitive_type = _types.try_primitive_type
map_descendability = _types.map_descendability
collect_ids_of_our_types_in_properties = _types.collect_ids_of_our_types_in_properties

translate = _translate.translate
errors_if_contracts_for_functions_or_methods_defined = (
    _translate.errors_if_contracts_for_functions_or_methods_defined
)
errors_if_non_implementation_specific_methods = (
    _translate.errors_if_non_implementation_specific_methods
)

dump = _stringify.dump
stringify = _stringify.stringify



"""Transpile Python to Go code."""
import abc
from typing import (
    Tuple,
    Optional,
    List,
    Mapping,
    Union,
    Set,
)

from icontract import ensure

from aas_core_codegen import intermediate
from aas_core_codegen.common import (
    Error,
    Stripped,
    assert_never,
    Identifier,
    indent_but_first_line,
)
from aas_core_codegen.golang import (
    common as golang_common,
    naming as golang_naming,
    pointering as golang_pointering,
)
from aas_core_codegen.golang.common import (
    INDENT as I,
    INDENT2 as II,
    INDENT3 as III,
)
from aas_core_codegen.intermediate import type_inference as intermediate_type_inference
from aas_core_codegen.parse import tree as parse_tree

# NOTE (mristin, 2023-06-01):
# We have to implement a very similar function for generating type annotations to
# aas_core_codegen.golang.common.generate_type since we can not simply pass
# intermediate_type_inference.TypeAnnotationUnion to
# aas_core_codegen.golang.common.generate_type.

PRIMITIVE_TYPE_MAP = {
    intermediate_type_inference.PrimitiveType.BOOL: Stripped("bool"),
    intermediate_type_inference.PrimitiveType.INT: Stripped("int64"),
    intermediate_type_inference.PrimitiveType.FLOAT: Stripped("float64"),
    intermediate_type_inference.PrimitiveType.STR: Stripped("string"),
    intermediate_type_inference.PrimitiveType.BYTEARRAY: Stripped("[]byte"),
    intermediate_type_inference.PrimitiveType.NONE: Stripped("struct{}"),
    intermediate_type_inference.PrimitiveType.LENGTH: Stripped("int"),
}


@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
def generate_type(
    type_annotation: intermediate_type_inference.TypeAnnotationUnion,
    types_package: Optional[Identifier] = None,
) -> Tuple[Optional[Stripped], Optional[str]]:
    """
    Generate the Go type for the given type annotation.

    If ``types_package`` is specified, it is prepended to all our types.

    (mristin, 2023-06-01): We do not handle all the type annotations from
    :py:mod:`aas_core_codegen.intermediate.type_inference` as that would be
    YAGNI (*e.g.*, verification functions, built-in functions *etc.*).
    If we do not know how to generate the type in Go, we return an error message.
    """
    if isinstance(type_annotation, intermediate_type_inference.PrimitiveTypeAnnotation):
        return PRIMITIVE_TYPE_MAP[type_annotation.a_type], None

    elif isinstance(type_annotation, intermediate_type_inference.OurTypeAnnotation):
        our_type = type_annotation.our_type

        if isinstance(our_type, intermediate.Enumeration):
            enum_name = golang_naming.enum_name(type_annotation.our_type.name)
            if types_package is None:
                return enum_name, None

            return Stripped(f"{types_package}.{enum_name}"), None

        elif isinstance(our_type, intermediate.ConstrainedPrimitive):
            return golang_common.PRIMITIVE_TYPE_MAP[our_type.constrainee], None

        elif isinstance(
            our_type, (intermediate.AbstractClass, intermediate.ConcreteClass)
        ):
            # NOTE (mristin, 2023-03-28):
            # We always refer to interfaces even in cases of concrete classes without
            # concrete descendants since we want to allow enhancing.
            interface_name = golang_naming.interface_name(our_type.name)

            if types_package is None:
                return interface_name, None

            return Stripped(f"{types_package}.{interface_name}"), None

    elif isinstance(type_annotation, intermediate_type_inference.ListTypeAnnotation):
        item_type, error_msg = generate_type(
            type_annotation=type_annotation.items, types_package=types_package
        )

        if error_msg is not None:
            return None, error_msg

        assert item_type is not None

        return Stripped(f"[]{item_type}"), None

    elif isinstance(
        type_annotation, intermediate_type_inference.OptionalTypeAnnotation
    ):
        value_type, error_msg = generate_type(
            type_annotation=type_annotation.value, types_package=types_package
        )

        if error_msg is not None:
            return None, error_msg

        assert value_type is not None

        if golang_pointering.is_pointer_type(type_annotation):
            return Stripped(f"*{value_type}"), None

        return value_type, None

    else:
        return None, (
            f"(mristin, 2023-06-01): We do not handle "
            f"the type annotation {type_annotation} from "
            "aas_core_codegen.intermediate.type_inference as that was, "
            "at this time point, YAGNI (*e.g.*, verification functions, "
            "built-in functions *etc.*). If you need this feature, please "
            "contact the developers."
        )

    raise AssertionError("Should not have gotten here")


class Transpiler(
    parse_tree.RestrictedTransformer[Tuple[Optional[Stripped], Optional[Error]]]
):
    """Transpile a node of our AST to Go code, or return an error."""

    _GOLANG_COMPARISON_MAP = {
        parse_tree.Comparator.LT: "<",
        parse_tree.Comparator.LE: "<=",
        parse_tree.Comparator.GT: ">",
        parse_tree.Comparator.GE: ">=",
        parse_tree.Comparator.EQ: "==",
        parse_tree.Comparator.NE: "!=",
    }

    def __init__(
        self,
        type_map: Mapping[
            parse_tree.Node, intermediate_type_inference.TypeAnnotationUnion
        ],
        is_pointer_map: Mapping[parse_tree.Node, bool],
        environment: intermediate_type_inference.Environment,
        types_package: Optional[Identifier] = None,
    ) -> None:
        """
        Initialize with the given values.

        If ``types_package`` is specified, it is prepended to all our types.
        """
        self.type_map = type_map
        self._is_pointer_map = is_pointer_map
        self._environment = intermediate_type_inference.MutableEnvironment(
            parent=environment
        )
        self._types_package = types_package

        # Keep track whenever we define a variable name, so that we can know how to
        # generate the reference in the Go code.
        self._variable_name_set = set()  # type: Set[Identifier]

    @ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
    def _transform_and_dereference_if_necessary(
        self, node: parse_tree.Node
    ) -> Tuple[Optional[Stripped], Optional[Error]]:
        """
        Dereference the given node if it is annotated as an optional.

        If the value denoted by ``node`` is not an optional, or does not need
        dereferencing, it is returned transpiled as-is.
        """
        if not isinstance(node, (parse_tree.Name, parse_tree.Member, parse_tree.Index)):
            return self.transform(node)

        code, error = self.transform(node)
        if error is not None:
            return None, error

        needs_dereferencing = self._is_pointer_map.get(node, None)
        if needs_dereferencing is None:
            error = Error(
                node.original_node,
                f"A node in our AST has not been mapped for "
                f"is-pointer: {parse_tree.dump(node)}; this is an assertion violation!",
            )
            return None, error

        if not needs_dereferencing:
            return code, None

        return Stripped(f"*{code}"), None

    @abc.abstractmethod
    def _transform_enumeration_literal(
        self, enumeration_name: Identifier, literal_name: Identifier
    ) -> Stripped:
        """
        Generate the code to represent an enumeration literal.

        In Go, enumeration literals are mere constants. Hence, we can not
        "de-reference" the enumeration literals from an enumeration, but
        generate the constant name here.
        """
        raise NotImplementedError()

    @ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
    def transform_member(
        self, node: parse_tree.Member
    ) -> Tuple[Optional[Stripped], Optional[Error]]:
        instance, error = self.transform(node.instance)
        if error is not None:
            return None, error

        # NOTE (mristin, 2023-04-12):
        # Ignore optional instance as they need to be checked before in the code
        instance_type = intermediate_type_inference.beneath_optional(
            self.type_map[node.instance]
        )

        # NOTE (mristin, 2023-05-16):
        # We explicitly do *not* dereference member access. Make sure you use
        # :py:meth:`_transform_and_dereference_if_necessary` where appropriate. Notably,
        # the operators ``is None`` and ``is not None`` have to compare against
        # references instead of values. That is why we can not simply de-reference
        # all the member access.

        member_type = self.type_map[node]

        member_accessor: str

        if isinstance(
            instance_type, intermediate_type_inference.OurTypeAnnotation
        ) and isinstance(instance_type.our_type, intermediate.Enumeration):
            # NOTE (mristin, 2023-01-13):
            # This member denotes an enumeration literal of an enumeration.
            # In Go, enumeration literals are mere constants. Hence, we can not
            # "de-reference" the enumeration literals from an enumeration, but
            # generate the constant name here.
            return (
                self._transform_enumeration_literal(
                    enumeration_name=instance_type.our_type.name, literal_name=node.name
                ),
                None,
            )

        elif isinstance(member_type, intermediate_type_inference.MethodTypeAnnotation):
            member_accessor = golang_naming.method_name(node.name)

        elif isinstance(
            instance_type, intermediate_type_inference.OurTypeAnnotation
        ) and isinstance(instance_type.our_type, intermediate.Class):
            if node.name in instance_type.our_type.properties_by_name:
                getter_name = golang_naming.getter_name(node.name)
                member_accessor = f"{getter_name}()"
            else:
                return None, Error(
                    node.original_node,
                    f"The property {node.name!r} has not been defined "
                    f"in the class {instance_type.our_type.name!r}",
                )

        elif isinstance(
            instance_type, intermediate_type_inference.EnumerationAsTypeTypeAnnotation
        ):
            if node.name in instance_type.enumeration.literals_by_name:
                # NOTE (mristin, 2023-01-13):
                # The member denotes an enumeration literal of an enumeration.
                # In Go, enumeration literals are mere constants. Hence, we can not
                # "de-reference" the enumeration literals from an enumeration, but
                # generate the constant name here.
                return (
                    self._transform_enumeration_literal(
                        enumeration_name=instance_type.enumeration.name,
                        literal_name=node.name,
                    ),
                    None,
                )
            else:
                return None, Error(
                    node.original_node,
                    f"The literal {node.name!r} has not been defined "
                    f"in the enumeration {instance_type.enumeration.name!r}",
                )
        else:
            return None, Error(
                node.original_node,
                f"We do not know how to generate the member access. The inferred type "
                f"of the instance was {instance_type}, while the member type "
                f"was {member_type}. However, we do not know how to resolve "
                f"the member {node.name!r} in {instance_type}.",
            )

        return Stripped(f"{instance}.{member_accessor}"), None

    @ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
    def transform_index(
        self, node: parse_tree.Index
    ) -> Tuple[Optional[Stripped], Optional[Error]]:
        collection, error = self.transform(node.collection)
        if error is not None:
            return None, error
        assert collection is not None

        index, error = self.transform(node.index)
        if error is not None:
            return None, error
        assert index is not None

        index_as_int = None  # type: Optional[int]
        try:
            index_as_int = int(index)
        except ValueError:
            pass

        if index_as_int is not None and index_as_int < 0:
            if "\n" in collection:
                # pylint: disable=invalid-unary-operand-type
                index = Stripped(
                    f"""\
len(
{I}{indent_but_first_line(collection, I)}
) - {-index_as_int}"""
                )
            else:
                index = Stripped(
                    f"len({collection}) - {-index_as_int}"  # pylint: disable=invalid-unary-operand-type
                )

        no_parentheses_types = (
            parse_tree.Member,
            parse_tree.FunctionCall,
            parse_tree.MethodCall,
            parse_tree.Name,
            parse_tree.Constant,
            parse_tree.Index,
            parse_tree.IsIn,
        )

        if not isinstance(node.collection, no_parentheses_types):
            collection = Stripped(f"({collection})")

        # NOTE (mristin, 2023-05-16):
        # We explicitly do *not* dereference index access. Make sure you use
        # :py:meth:`_transform_and_dereference_if_necessary` where appropriate. Notably,
        # the operators ``is None`` and ``is not None`` have to compare against
        # references instead of values. That is why we can not simply de-reference
        # all the index access.

        return Stripped(f"{collection}[{index}]"), None

    @ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
    def transform_comparison(
        self, node: parse_tree.Comparison
    ) -> Tuple[Optional[Stripped], Optional[Error]]:
        comparator = Transpiler._GOLANG_COMPARISON_MAP[node.op]

        errors = []

        left, error = self._transform_and_dereference_if_necessary(node.left)
        if error is not None:
            errors.append(error)

        right, error = self._transform_and_dereference_if_necessary(node.right)
        if error is not None:
            errors.append(error)

        if len(errors) > 0:
            return None, Error(
                node.original_node, "Failed to transpile the comparison", errors
            )

        no_parentheses_types = (
            parse_tree.Member,
            parse_tree.FunctionCall,
            parse_tree.MethodCall,
            parse_tree.Name,
            parse_tree.Constant,
            parse_tree.IsIn,
            parse_tree.Index,
            parse_tree.All,
            parse_tree.Any,
        )

        if not isinstance(node.left, no_parentheses_types):
            left = Stripped(f"({left})")

        if not isinstance(node.right, no_parentheses_types):
            right = Stripped(f"({right})")

        return Stripped(f"{left} {comparator} {right}"), None

    @ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
    def transform_is_in(
        self, node: parse_tree.IsIn
    ) -> Tuple[Optional[Stripped], Optional[Error]]:
        errors = []

        member, error = self._transform_and_dereference_if_necessary(node.member)
        if error is not None:
            errors.append(error)

        container, error = self._transform_and_dereference_if_necessary(node.container)
        if error is not None:
            errors.append(error)

        if len(errors) > 0:
            return None, Error(
                node.original_node,
                "Failed to transpile the membership relation",
                errors,
            )

        assert container is not None
        assert member is not None

        container_type = self.type_map[node.container]

        if isinstance(container_type, intermediate_type_inference.SetTypeAnnotation):
            return (
                Stripped(
                    f"""\
aascommon.MapContains(
{I}{indent_but_first_line(container, I)},
{I}{indent_but_first_line(member, I)},
)"""
                ),
                None,
            )

        else:
            return None, Error(
                node.original_node,
                f"We do not know how to generate the is-in operation for "
                f"the container {node.container}. The inferred type of "
                f"the container was {container_type}. "
                f"Please contact the developers if you need this feature.",
            )

    @ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
    def transform_implication(
        self, node: parse_tree.Implication
    ) -> Tuple[Optional[Stripped], Optional[Error]]:
        errors = []

        antecedent, error = self._transform_and_dereference_if_necessary(
            node.antecedent
        )
        if error is not None:
            errors.append(error)

        consequent, error = self._transform_and_dereference_if_necessary(
            node.consequent
        )
        if error is not None:
            errors.append(error)

        if len(errors) > 0:
            return None, Error(
                node.original_node, "Failed to transpile the implication", errors
            )

        assert antecedent is not None
        assert consequent is not None

        no_parentheses_types_in_this_context = (
            parse_tree.Member,
            parse_tree.FunctionCall,
            parse_tree.MethodCall,
            parse_tree.Name,
            parse_tree.IsIn,
            parse_tree.Index,
            parse_tree.All,
            parse_tree.Any,
        )

        if isinstance(node.antecedent, no_parentheses_types_in_this_context):
            not_antecedent = f"!{antecedent}"
        else:
            not_antecedent = f"!({antecedent})"

        if not isinstance(node.consequent, no_parentheses_types_in_this_context):
            consequent = Stripped(f"({consequent})")

        return Stripped(f"{not_antecedent} ||\n{consequent}"), None

    @ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
    def transform_method_call(
        self, node: parse_tree.MethodCall
    ) -> Tuple[Optional[Stripped], Optional[Error]]:
        errors = []  # type: List[Error]

        instance, error = self._transform_and_dereference_if_necessary(
            node.member.instance
        )
        if error is not None:
            errors.append(error)

        args = []  # type: List[Stripped]
        for arg_node in node.args:
            arg, error = self.transform(arg_node)
            if error is not None:
                errors.append(error)
                continue

            assert arg is not None

            args.append(arg)

        if len(errors) > 0:
            return None, Error(
                node.original_node, "Failed to transpile the method call", errors
            )

        assert instance is not None

        if not isinstance(node.member.instance, (parse_tree.Name, parse_tree.Member)):
            instance = Stripped(f"({instance})")

        method_name = golang_naming.method_name(node.member.name)

        args_joined = ", ".join(args)

        # Apply heuristic for breaking the lines
        if len(args_joined) > 50:
            args_joined = "\n".join(f"{arg}," for arg in args)

            return (
                Stripped(
                    f"""\
{instance}.{method_name}(
{I}{indent_but_first_line(args_joined, I)}
)"""
                ),
                None,
            )
        else:
            return Stripped(f"{instance}.{method_name}({args_joined})"), None

    @ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
    def transform_function_call(
        self, node: parse_tree.FunctionCall
    ) -> Tuple[Optional[Stripped], Optional[Error]]:
        errors = []  # type: List[Error]

        args = []  # type: List[Stripped]
        for arg_node in node.args:
            arg, error = self._transform_and_dereference_if_necessary(arg_node)
            if error is not None:
                errors.append(error)
                continue

            assert arg is not None

            args.append(arg)

        if len(errors) > 0:
            return None, Error(
                node.original_node, "Failed to transpile the function call", errors
            )

        # NOTE (mristin, 2023-03-28):
        # The validity of the arguments is checked in
        # :py:func:`aas_core_codegen.intermediate._translate.translate`, so we do not
        # have to test for argument arity here.

        func_type = self.type_map[node.name]

        if not isinstance(
            func_type, intermediate_type_inference.FunctionTypeAnnotationUnionAsTuple
        ):
            return None, Error(
                node.name.original_node,
                f"Expected the name to refer to a function, "
                f"but its inferred type was {func_type}",
            )

        if isinstance(
            func_type, intermediate_type_inference.VerificationTypeAnnotation
        ):
            function_name, error = self.transform_name(node.name)
            if error is not None:
                return None, error

            assert function_name is not None

            args_joined = ", ".join(args)

            # Apply heuristic for breaking the lines
            if len(function_name) + len(args_joined) > 50:
                args_joined = "\n".join(f"{arg}," for arg in args)
                return (
                    Stripped(
                        f"""\
{function_name}(
{I}{indent_but_first_line(args_joined, I)}
)"""
                    ),
                    None,
                )
            else:
                return Stripped(f"{function_name}({args_joined})"), None

        elif isinstance(
            func_type, intermediate_type_inference.BuiltinFunctionTypeAnnotation
        ):
            if func_type.func.name == "len":
                assert len(args) == 1, (
                    f"Expected exactly one argument, but got: {args}; "
                    f"this should have been caught before."
                )

                if "\n" in args[0]:
                    return (
                        Stripped(
                            f"""\
len(
{I}{indent_but_first_line(args[0], I)},
)"""
                        ),
                        None,
                    )

                return Stripped(f"len({args[0]})"), None

            else:
                return None, Error(
                    node.original_node,
                    f"The handling of the built-in function {node.name!r} has not "
                    f"been implemented",
                )
        else:
            assert_never(func_type)

        raise AssertionError("Should not have gotten here")

    def transform_constant(
        self, node: parse_tree.Constant
    ) -> Tuple[Optional[Stripped], Optional[Error]]:
        if isinstance(node.value, bool):
            return Stripped("true" if node.value else "false"), None
        elif isinstance(node.value, (int, float)):
            return Stripped(str(node.value)), None
        elif isinstance(node.value, str):
            return Stripped(golang_common.string_literal(node.value)), None
        else:
            assert_never(node.value)

        raise AssertionError("Should not have gotten here")

    def transform_is_none(
        self, node: parse_tree.IsNone
    ) -> Tuple[Optional[Stripped], Optional[Error]]:
        # NOTE (mristin, 2023-05-16):
        # We explicitly do not call :py:meth:`_transform_and_dereference_if_necessary`
        # here as we have to work on the pointer, not the value.

        value, error = self.transform(node.value)
        if error is not None:
            return None, error

        no_parentheses_types = (
            parse_tree.Name,
            parse_tree.Member,
            parse_tree.MethodCall,
            parse_tree.FunctionCall,
            parse_tree.IsIn,
            parse_tree.Index,
            parse_tree.All,
            parse_tree.Any,
        )
        if isinstance(node.value, no_parentheses_types):
            return Stripped(f"{value} == nil"), None
        else:
            return Stripped(f"({value}) == nil"), None

    def transform_is_not_none(
        self, node: parse_tree.IsNotNone
    ) -> Tuple[Optional[Stripped], Optional[Error]]:
        value, error = self.transform(node.value)
        if error is not None:
            return None, error

        no_parentheses_types_in_this_context = (
            parse_tree.Name,
            parse_tree.Member,
            parse_tree.MethodCall,
            parse_tree.FunctionCall,
            parse_tree.IsIn,
            parse_tree.Index,
            parse_tree.All,
            parse_tree.Any,
        )
        if isinstance(node.value, no_parentheses_types_in_this_context):
            return Stripped(f"{value} != nil"), None
        else:
            return Stripped(f"({value}) != nil"), None

    @abc.abstractmethod
    def transform_name(
        self, node: parse_tree.Name
    ) -> Tuple[Optional[Stripped], Optional[Error]]:
        raise NotImplementedError()

    def transform_not(
        self, node: parse_tree.Not
    ) -> Tuple[Optional[Stripped], Optional[Error]]:
        operand, error = self._transform_and_dereference_if_necessary(node.operand)
        if error is not None:
            return None, error

        no_parentheses_types_in_this_context = (
            parse_tree.Name,
            parse_tree.Member,
            parse_tree.MethodCall,
            parse_tree.FunctionCall,
            parse_tree.IsIn,
            parse_tree.Index,
            parse_tree.All,
            parse_tree.Any,
        )
        if not isinstance(node.operand, no_parentheses_types_in_this_context):
            return Stripped(f"!({operand})"), None
        else:
            return Stripped(f"!{operand}"), None

    def transform_and(
        self, node: parse_tree.And
    ) -> Tuple[Optional[Stripped], Optional[Error]]:
        errors = []  # type: List[Error]
        values = []  # type: List[Stripped]

        for value_node in node.values:
            value, error = self._transform_and_dereference_if_necessary(value_node)
            if error is not None:
                errors.append(error)
                continue

            assert value is not None

            no_parentheses_types_in_this_context = (
                parse_tree.Member,
                parse_tree.MethodCall,
                parse_tree.FunctionCall,
                parse_tree.Comparison,
                parse_tree.Name,
                parse_tree.IsIn,
                parse_tree.Index,
                parse_tree.All,
                parse_tree.Any,
            )

            if not isinstance(value_node, no_parentheses_types_in_this_context):
                value = Stripped(f"({value})")

            values.append(value)

        if len(errors) > 0:
            return None, Error(
                node.original_node, "Failed to transpile the conjunction", errors
            )

        values_joined = " &&\n".join(values)
        return Stripped(values_joined), None

    def transform_or(
        self, node: parse_tree.Or
    ) -> Tuple[Optional[Stripped], Optional[Error]]:
        errors = []  # type: List[Error]
        values = []  # type: List[Stripped]

        for value_node in node.values:
            value, error = self._transform_and_dereference_if_necessary(value_node)
            if error is not None:
                errors.append(error)
                continue

            assert value is not None

            no_parentheses_types_in_this_context = (
                parse_tree.Member,
                parse_tree.MethodCall,
                parse_tree.FunctionCall,
                parse_tree.Comparison,
                parse_tree.Name,
                parse_tree.IsIn,
                parse_tree.Index,
                parse_tree.All,
                parse_tree.Any,
            )

            if not isinstance(value_node, no_parentheses_types_in_this_context):
                value = Stripped(f"({value})")

            values.append(value)

        if len(errors) > 0:
            return None, Error(
                node.original_node, "Failed to transpile the conjunction", errors
            )

        values_joined = " ||\n".join(values)
        return Stripped(values_joined), None

    def _transform_add_or_sub(
        self, node: Union[parse_tree.Add, parse_tree.Sub]
    ) -> Tuple[Optional[Stripped], Optional[Error]]:
        errors = []  # type: List[Error]

        left, error = self._transform_and_dereference_if_necessary(node.left)
        if error is not None:
            errors.append(error)

        right, error = self._transform_and_dereference_if_necessary(node.right)
        if error is not None:
            errors.append(error)

        if len(errors) > 0:
            operation_name: str
            if isinstance(node, parse_tree.Add):
                operation_name = "the addition"
            elif isinstance(node, parse_tree.Sub):
                operation_name = "the subtraction"
            else:
                assert_never(node)

            return None, Error(
                node.original_node, f"Failed to transpile {operation_name}", errors
            )

        no_parentheses_types_in_this_context = (
            parse_tree.Member,
            parse_tree.MethodCall,
            parse_tree.FunctionCall,
            parse_tree.Constant,
            parse_tree.Name,
            parse_tree.IsIn,
            parse_tree.Index,
            parse_tree.All,
            parse_tree.Any,
        )

        if not isinstance(node.left, no_parentheses_types_in_this_context):
            left = Stripped(f"({left})")

        if not isinstance(node.right, no_parentheses_types_in_this_context):
            right = Stripped(f"({right})")

        if isinstance(node, parse_tree.Add):
            return Stripped(f"{left} + {right}"), None
        elif isinstance(node, parse_tree.Sub):
            return Stripped(f"{left} - {right}"), None
        else:
            assert_never(node)

    def transform_add(
        self, node: parse_tree.Add
    ) -> Tuple[Optional[Stripped], Optional[Error]]:
        return self._transform_add_or_sub(node)

    def transform_sub(
        self, node: parse_tree.Sub
    ) -> Tuple[Optional[Stripped], Optional[Error]]:
        return self._transform_add_or_sub(node)

    def transform_joined_str(
        self, node: parse_tree.JoinedStr
    ) -> Tuple[Optional[Stripped], Optional[Error]]:
        if all(isinstance(value, str) for value in node.values):
            text = "".join(node.values)  # type: ignore
            return golang_common.string_literal(text), None

        # NOTE (mristin, 2023-03-28):
        # We need the interpolation if we got so far.

        text_parts = []  # type: List[str]
        args = []  # type: List[str]

        for value in node.values:
            if isinstance(value, str):
                string_literal = golang_common.string_literal(value.replace("%", "%%"))

                # We need to remove double-quotes since we are joining everything
                # ourselves later.

                assert string_literal.startswith('"') and string_literal.endswith('"')

                string_literal_wo_quotes = string_literal[1:-1]
                text_parts.append(string_literal_wo_quotes)

            elif isinstance(value, parse_tree.FormattedValue):
                code, error = self._transform_and_dereference_if_necessary(value.value)
                if error is not None:
                    return None, error

                assert code is not None

                text_parts.append("%v")
                args.append(code)
            else:
                assert_never(value)

        string_literal = golang_common.string_literal("".join(text_parts))

        args_joined = "\n".join(f"{arg}," for arg in args)

        return (
            Stripped(
                f"""\
fmt.Sprintf(
{I}{indent_but_first_line(string_literal, I)},
{I}{indent_but_first_line(args_joined, I)},
)"""
            ),
            None,
        )

    def _transform_any_or_all(
        self, node: Union[parse_tree.Any, parse_tree.All]
    ) -> Tuple[Optional[Stripped], Optional[Error]]:
        errors = []  # type: List[Error]

        iteration = None  # type: Optional[Stripped]
        start = None  # type: Optional[Stripped]
        end = None  # type: Optional[Stripped]

        if isinstance(node.generator, parse_tree.ForEach):
            iteration, error = self.transform(node.generator.iteration)
            if error is not None:
                errors.append(error)

        elif isinstance(node.generator, parse_tree.ForRange):
            start, error = self._transform_and_dereference_if_necessary(
                node.generator.start
            )
            if error is not None:
                errors.append(error)

            end, error = self._transform_and_dereference_if_necessary(
                node.generator.end
            )
            if error is not None:
                errors.append(error)

        else:
            assert_never(node.generator)

        variable_name = node.generator.variable.identifier
        variable_type_annotation = self.type_map[node.generator.variable]

        variable_name_go = golang_naming.variable_name(variable_name)
        variable_type_go, error_msg = generate_type(
            type_annotation=variable_type_annotation, types_package=self._types_package
        )
        if error_msg is not None:
            errors.append(Error(node.generator.variable.original_node, error_msg))

        try:
            self._environment.set(
                identifier=variable_name, type_annotation=variable_type_annotation
            )
            self._variable_name_set.add(variable_name)

            condition, error = self._transform_and_dereference_if_necessary(
                node.condition
            )
            if error is not None:
                errors.append(error)

            variable, error = self._transform_and_dereference_if_necessary(
                node.generator.variable
            )
            if error is not None:
                errors.append(error)

        finally:
            self._variable_name_set.remove(variable_name)
            self._environment.remove(variable_name)

        if len(errors) > 0:
            return None, Error(
                node.original_node,
                "Failed to transpile the generator expression",
                errors,
            )

        assert (iteration is not None) ^ (start is not None and end is not None)

        assert variable is not None
        assert condition is not None

        if isinstance(node.generator, parse_tree.ForEach):
            assert iteration is not None

            qualifier_function: str
            if isinstance(node, parse_tree.Any):
                qualifier_function = "Some"
            elif isinstance(node, parse_tree.All):
                qualifier_function = "All"
            else:
                assert_never(node)

            assert qualifier_function is not None

            no_parentheses_types_in_this_context = (
                parse_tree.Member,
                parse_tree.MethodCall,
                parse_tree.FunctionCall,
                parse_tree.Name,
                parse_tree.IsIn,
                parse_tree.Index,
                parse_tree.All,
                parse_tree.Any,
            )

            if not isinstance(
                node.generator.iteration, no_parentheses_types_in_this_context
            ):
                source = Stripped(f"({iteration})")
            else:
                source = iteration

            return (
                Stripped(
                    f"""\
aascommon.{qualifier_function}(
{I}func({variable_name_go} {variable_type_go}) bool {{
{II}return {indent_but_first_line(condition, III)}
{I}}},
{I}{indent_but_first_line(source, I)},
)"""
                ),
                None,
            )

        elif isinstance(node.generator, parse_tree.ForRange):
            if isinstance(node, parse_tree.Any):
                qualifier_function = "SomeRange"
            elif isinstance(node, parse_tree.All):
                qualifier_function = "AllRange"
            else:
                assert_never(node)

            assert start is not None
            assert end is not None

            return (
                Stripped(
                    f"""\
aascommon.{qualifier_function}(
{I}func({variable_name_go} {variable_type_go}) bool {{
{II}return {indent_but_first_line(condition, III)}
{I}}},
{I}{indent_but_first_line(start, I)},
{I}{indent_but_first_line(end, I)},
)"""
                ),
                None,
            )

        else:
            assert_never(node.generator)

    def transform_any(
        self, node: parse_tree.Any
    ) -> Tuple[Optional[Stripped], Optional[Error]]:
        return self._transform_any_or_all(node)

    def transform_all(
        self, node: parse_tree.All
    ) -> Tuple[Optional[Stripped], Optional[Error]]:
        return self._transform_any_or_all(node)

    def transform_assignment(
        self, node: parse_tree.Assignment
    ) -> Tuple[Optional[Stripped], Optional[Error]]:
        errors = []  # type: List[Error]

        value, error = self._transform_and_dereference_if_necessary(node.value)
        if error is not None:
            errors.append(error)

        is_definition = False

        target = None  # type: Optional[Stripped]
        if isinstance(node.target, parse_tree.Name):
            type_anno = self._environment.find(identifier=node.target.identifier)
            if type_anno is None:
                # NOTE (mristin, 2023-06-23):
                # This is a variable definition as we did not specify the identifier
                # in the environment.

                is_definition = True

                type_anno = self.type_map[node.value]
                self._variable_name_set.add(node.target.identifier)
                self._environment.set(
                    identifier=node.target.identifier, type_annotation=type_anno
                )

                target, error = self.transform_name(node=node.target)
                if error is not None:
                    errors.append(error)
            else:
                target, error = self.transform(node=node.target)
                if error is not None:
                    errors.append(error)

        if len(errors) > 0:
            return None, Error(
                node.original_node, "Failed to transpile the assignment", errors
            )

        assert target is not None
        assert value is not None

        target_is_pointer = self._is_pointer_map[node.target]
        if target_is_pointer:
            target = Stripped(f"*{target}")

        assignment = "=" if not is_definition else ":="

        # NOTE (mristin, 2022-07-12):
        # This is a rudimentary heuristic for basic line breaks, but works well in
        # practice.
        if "\n" in value or len(value) > 50:
            return (
                Stripped(
                    f"""\
{target} {assignment}
{I}{indent_but_first_line(value, I)})"""
                ),
                None,
            )

        return Stripped(f"{target} {assignment} {value}"), None

    def transform_return(
        self, node: parse_tree.Return
    ) -> Tuple[Optional[Stripped], Optional[Error]]:
        if node.value is None:
            return Stripped("return"), None

        # NOTE (mristin, 2023-05-24):
        # This is a potential source of error. We infer the types based on nullability
        # checks, so the inferred type might be a non-nullable, but Golang pointers
        # remain pointers even after we check for them.
        #
        # The following transformation can not be resolved unless we know the explicit
        # return type that we are expected — if it is an optional, we should return
        # the value as-is, and if it is a non-optional we have to de-reference it.
        # For now, we leave it as-is, and will revisit this part of the code once
        # the meta-model requires it.

        value, error = self.transform(node.value)
        if error is not None:
            return None, error

        assert value is not None

        # NOTE (mristin, 2023-03-28):
        # This is a rudimentary heuristic for basic line breaks, but works well in
        # practice.
        if "\n" in value or len(value) > 50:
            return (
                Stripped(
                    f"""\
return {indent_but_first_line(value, I)}"""
                ),
                None,
            )

        return Stripped(f"return {value}"), None


# noinspection PyProtectedMember,PyProtectedMember
assert all(op in Transpiler._GOLANG_COMPARISON_MAP for op in parse_tree.Comparator)



"""Generate the code for conversion from and to Protocol Buffers."""

import io
import os
import pathlib
from typing import Tuple, Optional, List, TextIO

from icontract import ensure, require

import aas_core_codegen.python.common as python_common
import aas_core_codegen.python.description as python_description
import aas_core_codegen.python.naming as python_naming
import aas_core_codegen.python_protobuf
from aas_core_codegen import intermediate, run, specific_implementations
from aas_core_codegen.common import (
    Error,
    Stripped,
    assert_never,
    Identifier,
    indent_but_first_line,
)
from aas_core_codegen.python.common import INDENT as I, INDENT2 as II, INDENT3 as III
from aas_core_codegen.python_protobuf import naming as python_protobuf_naming

assert aas_core_codegen.python_protobuf.__doc__ == __doc__


# region From-protobuf


def _generate_from_pb_for_enum(
    enum: intermediate.Enumeration, aas_pb_module: python_common.QualifiedModuleName
) -> List[Stripped]:
    """
    Generate the code to convert an enum back from a protocol buffer.

    The ``aas_pb_module`` indicates the fully-qualified name of this base module.
    """
    literal_to_literal_items = []  # type: List[str]

    py_enum_name = python_naming.enum_name(enum.name)
    pb_enum_name = python_protobuf_naming.enum_name(enum.name)

    for literal in enum.literals:
        py_literal_name = python_naming.enum_literal_name(literal.name)
        pb_constant_name = python_protobuf_naming.enum_literal_constant_name(
            enumeration_name=enum.name, literal_name=literal.name
        )

        literal_to_literal_items.append(
            f"types_pb.{pb_enum_name}.{pb_constant_name}:\n"
            f"{I}types.{py_enum_name}.{py_literal_name}"
        )

    literal_to_literal_items_joined = ",\n".join(literal_to_literal_items)

    map_name = python_naming.private_constant_name(
        Identifier(f"{enum.name}_from_pb_map")
    )

    func_name = python_naming.function_name(Identifier(f"{enum.name}_from_pb"))

    if len(enum.literals) > 0:
        first_literal = enum.literals[0]
        first_literal_pb_constant = python_protobuf_naming.enum_literal_constant_name(
            enumeration_name=enum.name, literal_name=first_literal.name
        )
        first_literal_py_name = python_naming.enum_literal_name(first_literal.name)

        docstring = Stripped(
            f"""\
Parse ``that`` enum back from its Protocol Buffer representation.

>>> import {aas_pb_module}.types_pb2 as types_pb
>>> from {aas_pb_module}.pbization import {func_name}
>>> {func_name}(
...     types_pb.{pb_enum_name}.{first_literal_pb_constant}
... )
<{py_enum_name}.{first_literal_py_name}: {first_literal.value!r}>"""
        )
    else:
        docstring = Stripped(
            "Parse ``that`` enum back from its Protocol Buffer representation."
        )

    quoted_docstring = python_description.docstring(docstring)

    return [
        Stripped(
            f"""\
# fmt: off
{map_name} = {{
{I}{indent_but_first_line(literal_to_literal_items_joined, I)}
}}  # type: Mapping[types_pb.{pb_enum_name}, types.{py_enum_name}]
# fmt: on"""
        ),
        Stripped(
            f"""\
def {func_name}(
{I}that: types_pb.{pb_enum_name}
) -> types.{py_enum_name}:
{I}{indent_but_first_line(quoted_docstring, I)}
{I}return {map_name}[that]"""
        ),
    ]


def _determine_function_name_from_pb_for_class(
    cls: intermediate.ClassUnion,
) -> Identifier:
    """Determine from-pb function to use to parse the Protocol Buffer to a class."""
    if len(cls.concrete_descendants) > 0:
        convert_func_name = python_naming.function_name(
            Identifier(f"{cls.name}_from_pb_choice")
        )
    else:
        convert_func_name = python_naming.function_name(
            Identifier(f"{cls.name}_from_pb")
        )

    return convert_func_name


def _generate_concrete_from_pb_for_class(
    cls: intermediate.ConcreteClass, aas_pb_module: python_common.QualifiedModuleName
) -> Stripped:
    """
    Generate the function to convert an instance of ``cls`` back from a Protocol Buffer.

    This function performs explicitly no dispatch and directly converts back properties
    from the Protocol Buffer fields.

    The ``aas_pb_module`` indicates the fully-qualified name of this base module.
    """
    constructor_kwargs = []  # type: List[Stripped]

    assert set(cls.properties_by_name.keys()) == set(
        cls.constructor.arguments_by_name.keys()
    ), "(mristin) We assume that the properties match the constructor arguments."

    for prop in cls.properties:
        type_anno = intermediate.beneath_optional(prop.type_annotation)

        is_optional = isinstance(
            prop.type_annotation, intermediate.OptionalTypeAnnotation
        )

        pb_prop_name = python_protobuf_naming.property_name(prop.name)
        py_prop_name = python_naming.property_name(prop.name)

        primitive_type = intermediate.try_primitive_type(type_anno)
        if primitive_type is not None:
            if (
                primitive_type is intermediate.PrimitiveType.BOOL
                or primitive_type is intermediate.PrimitiveType.INT
                or primitive_type is intermediate.PrimitiveType.FLOAT
                or primitive_type is intermediate.PrimitiveType.STR
            ):
                constructor_kwargs.append(
                    Stripped(
                        f"""\
{py_prop_name}=(
{I}that.{pb_prop_name}
{I}if that.HasField({pb_prop_name!r})
{I}else None
)"""
                    )
                    if is_optional
                    else Stripped(f"{py_prop_name}=that.{pb_prop_name}")
                )
            elif primitive_type is intermediate.PrimitiveType.BYTEARRAY:
                constructor_kwargs.append(
                    Stripped(
                        f"""\
{py_prop_name}=(
{I}bytearray(that.{pb_prop_name})
{I}if that.HasField({pb_prop_name!r})
{I}else None
)"""
                    )
                    if is_optional
                    else Stripped(f"{py_prop_name}=bytearray(that.{pb_prop_name})")
                )
            else:
                assert_never(primitive_type)

        else:
            if isinstance(type_anno, intermediate.PrimitiveTypeAnnotation):
                raise AssertionError("Expected to be handled before")

            elif isinstance(type_anno, intermediate.OurTypeAnnotation):
                if isinstance(type_anno.our_type, intermediate.Enumeration):
                    convert_func_name = python_naming.function_name(
                        Identifier(f"{type_anno.our_type.name}_from_pb")
                    )

                    constructor_kwargs.append(
                        Stripped(
                            f"""\
{py_prop_name}=(
{I}{convert_func_name}(
{II}that.{pb_prop_name}
{I})
{I}if that.HasField({pb_prop_name!r})
{I}else None
)"""
                        )
                        if is_optional
                        else Stripped(
                            f"""\
{py_prop_name}={convert_func_name}(
{I}that.{pb_prop_name}
)"""
                        )
                    )

                elif isinstance(type_anno.our_type, intermediate.ConstrainedPrimitive):
                    raise AssertionError("Expected to be handled before")

                elif isinstance(
                    type_anno.our_type,
                    (intermediate.AbstractClass, intermediate.ConcreteClass),
                ):
                    convert_func_name = _determine_function_name_from_pb_for_class(
                        cls=type_anno.our_type
                    )

                    constructor_kwargs.append(
                        Stripped(
                            f"""\
{py_prop_name}=(
{I}{convert_func_name}(
{II}that.{pb_prop_name}
{I})
{I}if that.HasField({pb_prop_name!r})
{I}else None
)"""
                        )
                        if is_optional
                        else Stripped(
                            f"""\
{py_prop_name}={convert_func_name}(
{I}that.{pb_prop_name}
)"""
                        )
                    )

                else:
                    assert_never(type_anno.our_type)

            elif isinstance(type_anno, intermediate.ListTypeAnnotation):
                assert isinstance(
                    type_anno.items, intermediate.OurTypeAnnotation
                ) and isinstance(
                    type_anno.items.our_type,
                    (intermediate.AbstractClass, intermediate.ConcreteClass),
                ), (
                    f"NOTE (mristin): We expect only lists of classes "
                    f"at the moment, but you specified {type_anno}. "
                    f"Please contact the developers if you need this feature."
                )

                convert_func_name = _determine_function_name_from_pb_for_class(
                    cls=type_anno.items.our_type
                )

                # NOTE (mristin):
                # Protocol Buffers 3 do not support ``HasField`` on repeated fields,
                # see: https://github.com/protocolbuffers/protobuf/issues/10489
                #
                # We decide here to interpret empty repeated list containers as None
                # if the model property is optional, and as an empty list if the model
                # property is required.
                constructor_kwargs.append(
                    Stripped(
                        f"""\
{py_prop_name}=(
{I}list(map(
{II}{convert_func_name},
{II}that.{pb_prop_name}
{I}))
{I}if len(that.{pb_prop_name}) > 0
{I}else None
)"""
                    )
                    if is_optional
                    else Stripped(
                        f"""\
{py_prop_name}=list(map(
{I}{convert_func_name},
{I}that.{pb_prop_name}
))"""
                    )
                )
            else:
                assert_never(type_anno)

    func_name = python_naming.function_name(Identifier(f"{cls.name}_from_pb"))

    pb_cls_name = python_protobuf_naming.class_name(cls.name)
    py_cls_name = python_naming.class_name(cls.name)

    constructor_kwargs_joined = ",\n".join(constructor_kwargs)

    py_var = python_naming.variable_name(cls.name)
    pb_var = python_naming.variable_name(Identifier(cls.name + "_pb"))

    return Stripped(
        f'''\
def {func_name}(
{I}that: types_pb.{pb_cls_name}
) -> types.{py_cls_name}:
{I}"""
{I}Parse ``that`` Protocol Buffer to an instance of a concrete class.

{I}Example usage:

{I}.. code-block::

{II}import {aas_pb_module}.types_pb2 as types_pb
{II}from {aas_pb_module}.pbization import {func_name}

{II}some_bytes = b'... serialized types_pb.{pb_cls_name} ...'
{II}{pb_var} = types_pb.{pb_cls_name}()
{II}{pb_var}.FromString(
{III}some_bytes
{II})

{II}{py_var} = {func_name}(
{III}{pb_var}
{II})
{II}# Do something with the {py_var}...

{I}"""
{I}return types.{py_cls_name}(
{II}{indent_but_first_line(constructor_kwargs_joined, II)}
{I})'''
    )


@require(
    lambda cls: len(cls.concrete_descendants) > 0,
    "Dispatch only possible with concrete descendants.",
)
def _generate_from_pb_for_class_choice(
    cls: intermediate.ClassUnion, aas_pb_module: python_common.QualifiedModuleName
) -> List[Stripped]:
    """
    Generate the function to parse an instance of ``cls`` where a dispatch is necessary.

    If the class has concrete descendants, the generated function needs to determine
    the concrete runtime type of the Protocol Buffer, and dispatch it to
    the corresponding concrete from-pb function.

    The ``aas_pb_module`` indicates the fully-qualified name of this base module.
    """
    blocks = []  # type: List[Stripped]

    # region Dispatch map
    items = []  # type: List[Stripped]

    concrete_classes = []  # type: List[intermediate.ConcreteClass]
    if isinstance(cls, intermediate.ConcreteClass):
        concrete_classes.append(cls)

    concrete_classes.extend(cls.concrete_descendants)

    for concrete_cls in concrete_classes:
        pb_which_one_of = python_protobuf_naming.property_name(concrete_cls.name)

        concrete_from_pb_name = python_naming.function_name(
            Identifier(f"{concrete_cls.name}_from_pb")
        )

        property_name = python_protobuf_naming.property_name(concrete_cls.name)

        items.append(
            Stripped(
                f"""\
{pb_which_one_of!r}:
{I}lambda that: {concrete_from_pb_name}(
{II}that.{property_name}
{I})"""
            )
        )

    map_name = python_naming.private_constant_name(
        Identifier(f"{cls.name}_from_pb_choice_map")
    )

    items_joined = ",\n".join(items)
    blocks.append(
        Stripped(
            f"""\
# fmt: off
{map_name} = {{
{I}{indent_but_first_line(items_joined, I)}
}}
# fmt: on"""
        )
    )

    # endregion Dispatch map

    # region Function

    pb_cls_name = python_protobuf_naming.choice_class_name(cls.name)

    func_name = python_naming.function_name(Identifier(f"{cls.name}_from_pb_choice"))

    py_cls_name = python_naming.class_name(cls.name)

    pb_var = python_naming.variable_name(Identifier(cls.name + "_choice_pb"))
    py_var = python_naming.variable_name(cls.name)

    blocks.append(
        Stripped(
            f'''\
def {func_name}(
{I}that: types_pb.{pb_cls_name}
) -> types.{py_cls_name}:
{I}"""
{I}Parse ``that`` Protocol Buffer based on its runtime ``WhichOneof``.

{I}Example usage:

{I}.. code-block::

{II}import {aas_pb_module}.types_pb2 as types_pb
{II}from {aas_pb_module}.pbization import {func_name}

{II}some_bytes = b'... serialized types_pb.{pb_cls_name} ...'
{II}{pb_var} = types_pb.{pb_cls_name}()
{II}{pb_var}.FromString(
{III}some_bytes
{II})

{II}{py_var} = {func_name}(
{III}{pb_var}
{II})
{II}# Do something with the {py_var}...
{I}"""
{I}get_concrete_instance_from_pb = (
{II}{map_name}[
{III}that.WhichOneof("value")
{II}]
{I})

{I}result = get_concrete_instance_from_pb(that)  # type: ignore

{I}assert isinstance(result, types.{py_cls_name})
{I}return result'''
        )
    )

    return blocks


def _generate_general_from_pb(
    symbol_table: intermediate.SymbolTable,
    aas_pb_module: python_common.QualifiedModuleName,
) -> List[Stripped]:
    """
    Generate the parsing from a Protocol Buffer based on its runtime type.

    The ``aas_pb_module`` indicates the fully-qualified name of this base module.
    """
    map_items = []  # type: List[Stripped]
    for cls in symbol_table.classes:
        if len(cls.concrete_descendants) > 0:
            pb_cls_name = python_protobuf_naming.choice_class_name(cls.name)

            from_pb_func = python_naming.function_name(
                Identifier(f"{cls.name}_from_pb_choice")
            )
        else:
            pb_cls_name = python_protobuf_naming.class_name(cls.name)

            from_pb_func = python_naming.function_name(
                Identifier(f"{cls.name}_from_pb")
            )

        map_items.append(
            Stripped(
                f"""\
types_pb.{pb_cls_name}:
{I}{from_pb_func}"""
            )
        )

    map_items_joined = ",\n".join(map_items)

    # NOTE (mristin):
    # We had to put this message outside since black formatter struggled with it. This
    # is most probably a bug in the black formatter.
    key_class_not_found_message = '''\
f"We do not know how to parse the protocol buffer "
f"of type {that.__class__} into a model instance."'''

    first_cls_name = symbol_table.concrete_classes[0].name
    pb_first_cls_name = python_protobuf_naming.class_name(first_cls_name)

    return [
        Stripped(
            f"""\
# fmt: off
_FROM_PB_MAP = {{
{I}{indent_but_first_line(map_items_joined, I)}
}}
# fmt: on"""
        ),
        Stripped(
            f'''\
def from_pb(
{I}that: google.protobuf.message.Message
) -> types.Class:
{I}"""
{I}Parse ``that`` Protocol Buffer into a model instance.

{I}The concrete parsing is determined based on the runtime type of ``that``
{I}Protocol Buffer. It is assumed that ``that`` is an instance of a message
{I}coming from the Protocol Buffer definitions corresponding to the meta-model.

{I}Example usage:

{I}.. code-block::

{II}import {aas_pb_module}.types_pb2 as types_pb
{II}from {aas_pb_module}.pbization import from_pb

{II}some_bytes = b'... serialized types_pb.{pb_first_cls_name} ...'
{II}instance_pb = types_pb.{pb_first_cls_name}()
{II}instance_pb.FromString(
{III}some_bytes
{II})

{II}instance = from_pb(
{III}instance_pb
{II})
{II}# Do something with the instance...
{I}"""
{I}from_pb_func = _FROM_PB_MAP.get(that.__class__, None)

{I}if from_pb_func is None:
{II}raise ValueError(
{III}{indent_but_first_line(key_class_not_found_message, III)}
{II})

{I}result = from_pb_func(that)  # type: ignore
{I}assert isinstance(result, types.Class)
{I}return result'''
        ),
    ]


@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
def _generate_from_pb(
    symbol_table: intermediate.SymbolTable,
    aas_pb_module: python_common.QualifiedModuleName,
) -> Tuple[Optional[List[Stripped]], Optional[List[Error]]]:
    """
    Generate all the code for the conversion from protocol buffers.

    The ``aas_module`` indicates the fully-qualified name of the base SDK module.

    The ``aas_pb_module`` indicates the fully-qualified name of this base module.
    """
    blocks = []  # type: List[Stripped]
    errors = []  # type: List[Error]

    for our_type in symbol_table.our_types:
        if isinstance(our_type, intermediate.Enumeration):
            blocks.extend(
                _generate_from_pb_for_enum(enum=our_type, aas_pb_module=aas_pb_module)
            )

        elif isinstance(our_type, intermediate.ConstrainedPrimitive):
            # NOTE (mristin):
            # We do not represent constrained primitives in Protocol Buffer types.
            pass

        elif isinstance(
            our_type, (intermediate.AbstractClass, intermediate.ConcreteClass)
        ):
            if len(our_type.concrete_descendants) > 0:
                blocks.extend(
                    _generate_from_pb_for_class_choice(
                        cls=our_type, aas_pb_module=aas_pb_module
                    )
                )

            if isinstance(our_type, intermediate.ConcreteClass):
                blocks.append(
                    _generate_concrete_from_pb_for_class(
                        cls=our_type, aas_pb_module=aas_pb_module
                    )
                )

        else:
            assert_never(our_type)

    blocks.extend(
        _generate_general_from_pb(
            symbol_table=symbol_table, aas_pb_module=aas_pb_module
        )
    )

    if len(errors) > 0:
        return None, errors

    return blocks, None


# endregion


# region To-protobuf


def _generate_partial_visitor(symbol_table: intermediate.SymbolTable) -> Stripped:
    """Generate the transformer where all methods return assertion errors."""
    methods = []  # type: List[Stripped]

    for cls in symbol_table.concrete_classes:
        method_name = python_naming.method_name(
            Identifier(f"visit_{cls.name}_with_context")
        )

        cls_name = python_naming.class_name(cls.name)

        methods.append(
            Stripped(
                f"""\
def {method_name}(
{I}self,
{I}that: types.{cls_name},
{I}context: T
) -> None:
{I}raise AssertionError(f"Unexpected visitation of {{that.__class__}}")"""
            )
        )
    body = "\n\n".join(methods)
    return Stripped(
        f'''\
class _PartialVisitorWithContext(types.AbstractVisitorWithContext[T]):
{I}"""
{I}Visit instances in context with double-dispatch.

{I}This class is meant to be inherited from. If you do not override a method,
{I}it will raise an exception. This is a partial visitor, meaning that some
{I}visits are unexpected by design.
{I}"""
{I}# pylint: disable=missing-docstring

{I}{indent_but_first_line(body, I)}'''
    )


def _generate_to_pb_for_enum(
    enum: intermediate.Enumeration,
    aas_module: python_common.QualifiedModuleName,
    aas_pb_module: python_common.QualifiedModuleName,
) -> List[Stripped]:
    """
    Generate the code to convert an enum to a protocol buffer.

    The ``aas_module`` indicates the fully-qualified name of the base SDK module.

    The ``aas_pb_module`` indicates the fully-qualified name of this base module.
    """
    literal_to_literal_items = []  # type: List[str]

    py_enum_name = python_naming.enum_name(enum.name)
    pb_enum_name = python_protobuf_naming.enum_name(enum.name)

    for literal in enum.literals:
        py_literal_name = python_naming.enum_literal_name(literal.name)
        pb_constant_name = python_protobuf_naming.enum_literal_constant_name(
            enumeration_name=enum.name, literal_name=literal.name
        )

        literal_to_literal_items.append(
            f"types.{py_enum_name}.{py_literal_name}:\n"
            f"{I}types_pb.{pb_enum_name}.{pb_constant_name}"
        )

    literal_to_literal_items_joined = ",\n".join(literal_to_literal_items)

    map_name = python_naming.private_constant_name(Identifier(f"{enum.name}_to_pb_map"))

    func_name = python_naming.function_name(Identifier(f"{enum.name}_to_pb"))

    if len(enum.literals) > 0:
        first_literal = enum.literals[0]
        first_literal_py_name = python_naming.enum_literal_name(first_literal.name)

        docstring = Stripped(
            f"""\
Convert ``that`` enum to its Protocol Buffer representation.

>>> from {aas_pb_module}.pbization import {func_name}
>>> import {aas_module}.types as types
>>> {func_name}(
...     types.{py_enum_name}.{first_literal_py_name}
... )
1"""
        )
    else:
        docstring = Stripped(
            "Convert ``that`` enum to its Protocol Buffer representation."
        )

    quoted_docstring = python_description.docstring(docstring)

    return [
        Stripped(
            f"""\
# fmt: off
{map_name} = {{
{I}{indent_but_first_line(literal_to_literal_items_joined, I)}
}}  # type: Mapping[types.{py_enum_name}, types_pb.{pb_enum_name}]
# fmt: on"""
        ),
        Stripped(
            f"""\
def {func_name}(
{I}that: types.{py_enum_name}
) -> types_pb.{pb_enum_name}:
{I}{indent_but_first_line(quoted_docstring, I)}
{I}return {map_name}[that]"""
        ),
    ]


def _generate_concrete_to_pb_for_class(
    cls: intermediate.ConcreteClass,
    aas_module: python_common.QualifiedModuleName,
    aas_pb_module: python_common.QualifiedModuleName,
) -> Stripped:
    """
    Generate the code to convert an instance to a protocol buffer without dispatch.

    There are two situations when this generated function will be called:

    1) ``cls`` is a concrete class without descendants (no dispatch necessary), and
    2) ``cls`` is a concrete class with descendants. In this case, we will dispatch
       to this generated function if the runtime type equals exactly ``cls``.

    The ``aas_module`` indicates the fully-qualified name of the base SDK module.

    The ``aas_pb_module`` indicates the fully-qualified name of this base module.
    """
    set_blocks = []  # type: List[Stripped]

    for prop in cls.properties:
        type_anno = intermediate.beneath_optional(prop.type_annotation)

        py_prop_name = python_naming.property_name(prop.name)
        pb_prop_name = python_protobuf_naming.property_name(prop.name)

        set_block: Stripped

        primitive_type = intermediate.try_primitive_type(type_anno)
        if primitive_type is not None:
            if (
                primitive_type is intermediate.PrimitiveType.BOOL
                or primitive_type is intermediate.PrimitiveType.INT
                or primitive_type is intermediate.PrimitiveType.FLOAT
                or primitive_type is intermediate.PrimitiveType.STR
            ):
                set_block = Stripped(f"target.{pb_prop_name} = that.{py_prop_name}")

            elif primitive_type is intermediate.PrimitiveType.BYTEARRAY:
                set_block = Stripped(
                    f"target.{pb_prop_name} = bytes(that.{py_prop_name})"
                )

            else:
                assert_never(primitive_type)
        else:
            if isinstance(type_anno, intermediate.PrimitiveTypeAnnotation):
                raise AssertionError("Expected to be handled before")

            elif isinstance(type_anno, intermediate.OurTypeAnnotation):
                if isinstance(type_anno.our_type, intermediate.Enumeration):
                    to_pb_func_for_enum = python_naming.function_name(
                        Identifier(f"{type_anno.our_type.name}_to_pb")
                    )

                    set_block = Stripped(
                        f"""\
target.{pb_prop_name} = {to_pb_func_for_enum}(
{I}that.{py_prop_name}
)"""
                    )

                elif isinstance(type_anno.our_type, intermediate.ConstrainedPrimitive):
                    raise AssertionError("Expected to be handled before")

                elif isinstance(
                    type_anno.our_type,
                    (intermediate.AbstractClass, intermediate.ConcreteClass),
                ):
                    if len(type_anno.our_type.concrete_descendants) > 0:
                        to_pb_choice_func = python_naming.function_name(
                            Identifier(f"{type_anno.our_type.name}_to_pb_choice")
                        )

                        set_block = Stripped(
                            f"""\
{to_pb_choice_func}(
{I}that.{py_prop_name},
{I}target.{pb_prop_name}
)"""
                        )
                    else:
                        to_pb_func = python_naming.function_name(
                            Identifier(f"{type_anno.our_type.name}_to_pb")
                        )

                        set_block = Stripped(
                            f"""\
# We clear so that the field is set even if all the properties are None.
target.{pb_prop_name}.Clear()

{to_pb_func}(
{I}that.{py_prop_name},
{I}target.{pb_prop_name}
)"""
                        )

                else:
                    assert_never(type_anno.our_type)

            elif isinstance(type_anno, intermediate.ListTypeAnnotation):
                assert isinstance(
                    type_anno.items, intermediate.OurTypeAnnotation
                ) and isinstance(
                    type_anno.items.our_type,
                    (intermediate.AbstractClass, intermediate.ConcreteClass),
                ), (
                    f"NOTE (mristin): We expect only lists of classes "
                    f"at the moment, but you specified {type_anno}. "
                    f"Please contact the developers if you need this feature."
                )

                if len(type_anno.items.our_type.concrete_descendants) > 0:
                    to_pb_func_for_items_cls = python_naming.function_name(
                        Identifier(f"{type_anno.items.our_type.name}_to_pb_choice")
                    )
                else:
                    to_pb_func_for_items_cls = python_naming.function_name(
                        Identifier(f"{type_anno.items.our_type.name}_to_pb")
                    )

                item_var = python_naming.variable_name(Identifier(f"{prop.name}_item"))
                set_block = Stripped(
                    f"""\
for {item_var} in that.{py_prop_name}:
{I}{item_var}_pb = target.{pb_prop_name}.add()
{I}{to_pb_func_for_items_cls}(
{II}{item_var},
{II}{item_var}_pb)"""
                )

            else:
                assert_never(type_anno)

        if isinstance(prop.type_annotation, intermediate.OptionalTypeAnnotation):
            set_block = Stripped(
                f"""\
if that.{py_prop_name} is not None:
{I}{indent_but_first_line(set_block, I)}"""
            )

        set_blocks.append(set_block)

    func_name = python_naming.function_name(Identifier(f"{cls.name}_to_pb"))

    py_cls_name = python_naming.class_name(cls.name)
    pb_cls_name = python_protobuf_naming.class_name(cls.name)

    set_blocks_joined = "\n\n".join(set_blocks)

    pb_var = python_naming.variable_name(Identifier(cls.name + "_pb"))
    py_var = python_naming.variable_name(cls.name)

    return Stripped(
        f'''\
def {func_name}(
{I}that: types.{py_cls_name},
{I}target: types_pb.{pb_cls_name}
) -> None:
{I}"""
{I}Set fields in ``target`` based on ``that`` instance.

{I}Example usage:

{I}.. code-block::

{II}import {aas_module}.types as types

{II}import {aas_pb_module}.types_pb2 as types_pb
{II}from {aas_pb_module}.pbization import {func_name}

{II}{py_var} = types.{py_cls_name}(
{III}... # some constructor arguments
{II})

{II}{pb_var} = types_pb.{pb_cls_name}()
{II}{func_name}(
{III}{py_var},
{III}{pb_var}
{II})

{II}some_bytes = {pb_var}.SerializeToString()
{II}# Do something with some_bytes. For example, transfer them
{II}# over the wire.
{I}"""
{I}{indent_but_first_line(set_blocks_joined, I)}'''
    )


@require(
    lambda cls: len(cls.concrete_descendants) > 0,
    "Dispatch needs concrete descendants.",
)
def _generate_to_pb_choice(
    cls: intermediate.ClassUnion,
    aas_module: python_common.QualifiedModuleName,
    aas_pb_module: python_common.QualifiedModuleName,
) -> List[Stripped]:
    """
    Generate the code to dispatch the serialization to a concrete serializer.

    The ``aas_module`` indicates the fully-qualified name of the base SDK module.

    The ``aas_pb_module`` indicates the fully-qualified name of this base module.
    """
    blocks = []  # type: List[Stripped]

    pb_choice_cls_name = python_protobuf_naming.choice_class_name(cls.name)

    # region Visitor

    methods = []  # type: List[Stripped]

    concrete_classes = []  # type: List[intermediate.ConcreteClass]
    if isinstance(cls, intermediate.ConcreteClass):
        concrete_classes.append(cls)

    concrete_classes.extend(cls.concrete_descendants)

    for concrete_cls in concrete_classes:
        py_concrete_cls_name = python_naming.class_name(concrete_cls.name)
        pb_property = python_protobuf_naming.property_name(concrete_cls.name)

        conversion_func = python_naming.function_name(
            Identifier(f"{concrete_cls.name}_to_pb")
        )

        method_name = python_naming.method_name(
            Identifier(f"visit_{concrete_cls.name}_with_context")
        )

        methods.append(
            Stripped(
                f'''\
def {method_name}(
{I}self,
{I}that: types.{py_concrete_cls_name},
{I}context: types_pb.{pb_choice_cls_name}
) -> None:
{I}"""
{I}Set the fields of ``context.{pb_property}``
{I}according to ``that`` instance.
{I}"""
{I}{conversion_func}(
{II}that,
{II}context.{pb_property}
{I})'''
            )
        )

    visitor_class_name = python_naming.private_class_name(
        Identifier(f"{cls.name}_to_pb_choice")
    )
    body = "\n\n".join(methods)
    blocks.append(
        Stripped(
            f'''\
class {visitor_class_name}(
{I}_PartialVisitorWithContext[
{II}types_pb.{pb_choice_cls_name}
{I}]
):
{I}"""Set the fields of the corresponding one-of value."""
{I}{indent_but_first_line(body, I)}'''
        )
    )

    visitor_name = python_naming.private_constant_name(
        Identifier(f"{cls.name}_to_pb_choice")
    )
    blocks.append(Stripped(f"{visitor_name} = {visitor_class_name}()"))

    # endregion Visitor

    func_name = python_naming.function_name(Identifier(f"{cls.name}_to_pb_choice"))

    py_cls_name = python_naming.class_name(cls.name)

    py_concrete_descendant_cls_name = python_naming.class_name(
        cls.concrete_descendants[0].name
    )

    py_var = python_naming.variable_name(cls.name)
    pb_var = python_naming.variable_name(Identifier(cls.name + "_choice_pb"))

    blocks.append(
        Stripped(
            f'''\
def {func_name}(
{I}that: types.{py_cls_name},
{I}target: types_pb.{pb_choice_cls_name}
) -> None:
{I}"""
{I}Set the chosen value in ``target`` based on ``that`` instance.

{I}The chosen value in ``target`` is determined based on the runtime type of ``that``
{I}instance. All the fields of the value are recursively set according to ``that``
{I}instance.

{I}Example usage:

{I}.. code-block::

{II}import {aas_module}.types as types

{II}import {aas_pb_module}.types_pb2 as types_pb
{II}from {aas_pb_module}.pbization import {func_name}

{II}{py_var} = types.{py_concrete_descendant_cls_name}(
{III}... # some constructor arguments
{II})

{II}{pb_var} = types_pb.{pb_choice_cls_name}()
{II}{func_name}(
{III}{py_var},
{III}{pb_var}
{II})

{II}some_bytes = {pb_var}.SerializeToString()
{II}# Do something with some_bytes. For example, transfer them
{II}# over the wire.

{I}"""
{I}{visitor_name}.visit_with_context(
{II}that,
{II}target
{I})'''
        )
    )

    return blocks


@require(
    lambda symbol_table: len(symbol_table.concrete_classes) > 0,
    "At least one concrete class in the meta-model expected if you generate "
    "a general to-pb function",
)
def _generate_general_to_pb(
    symbol_table: intermediate.SymbolTable,
    aas_module: python_common.QualifiedModuleName,
    aas_pb_module: python_common.QualifiedModuleName,
) -> List[Stripped]:
    """
    Generate a dispatch to the concrete to-pb function.

    The ``aas_module`` indicates the fully-qualified name of the base SDK module.

    The ``aas_pb_module`` indicates the fully-qualified name of this base module.
    """
    methods = []  # type: List[Stripped]

    for cls in symbol_table.concrete_classes:
        if len(cls.concrete_descendants) > 0:
            conversion_func_name = python_naming.function_name(
                Identifier(f"{cls.name}_to_pb_choice")
            )
            pb_cls_name = python_protobuf_naming.choice_class_name(cls.name)
        else:
            conversion_func_name = python_naming.function_name(
                Identifier(f"{cls.name}_to_pb")
            )
            pb_cls_name = python_protobuf_naming.class_name(cls.name)

        method_name = python_naming.function_name(Identifier(f"transform_{cls.name}"))

        py_cls_name = python_naming.class_name(cls.name)

        methods.append(
            Stripped(
                f'''\
def {method_name}(
{I}self,
{I}that: types.{py_cls_name}
) -> google.protobuf.message.Message:
{I}"""
{I}Convert ``that`` instance
{I}to a :py:class:`types_pb.{pb_cls_name}`.
{I}"""
{I}result = types_pb.{pb_cls_name}()
{I}{conversion_func_name}(that, result)
{I}return result'''
            )
        )

    body = "\n\n".join(methods)

    first_cls = symbol_table.concrete_classes[0]
    py_first_cls_name = python_naming.class_name(first_cls.name)

    return [
        Stripped(
            f'''\
class _ToPbTransformer(
{I}types.AbstractTransformer[google.protobuf.message.Message]
):
{I}"""
{I}Dispatch to-pb conversion to the concrete functions.

{I}The classes with descendants (i.e., subtypes) are always going to be converted
{I}to their concrete Protocol Buffer instead of the choice (union) Protocol Buffer
{I}class. We made this decision with the compactness of messages in mind.
{I}"""
{I}{indent_but_first_line(body, I)}'''
        ),
        Stripped("_TO_PB_TRANSFORMER = _ToPbTransformer()"),
        Stripped(
            f'''\
def to_pb(
{I}that: types.Class,
) -> google.protobuf.message.Message:
{I}"""
{I}Dispatch to-pb conversion to the concrete functions.

{I}The classes with descendants (i.e., subtypes) are always going to be converted
{I}to their concrete Protocol Buffer message type instead of the choice (union) type.
{I}We made this decision with the compactness of messages in mind as choice types
{I}would occupy a tiny bit more space.

{I}Example usage:

{I}.. code-block::

{II}import {aas_module}.types as types

{II}from {aas_pb_module}.pbization import to_pb

{II}instance = types.{py_first_cls_name}(
{III}... # some constructor arguments
{II})

{II}instance_pb = to_pb(
{III}instance
{II})

{II}some_bytes = instance_pb.SerializeToString()
{II}# Do something with some_bytes. For example, transfer them
{II}# over the wire.
{I}"""
{I}return _TO_PB_TRANSFORMER.transform(that)'''
        ),
    ]


@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
def _generate_to_pb(
    symbol_table: intermediate.SymbolTable,
    aas_module: python_common.QualifiedModuleName,
    aas_pb_module: python_common.QualifiedModuleName,
) -> Tuple[Optional[List[Stripped]], Optional[List[Error]]]:
    """
    Generate all the code for the conversion to protocol buffers.

    The ``aas_module`` indicates the fully-qualified name of the base SDK module.

    The ``aas_pb_module`` indicates the fully-qualified name of this base module.
    """
    blocks = [
        Stripped('T = TypeVar("T")'),
        _generate_partial_visitor(symbol_table=symbol_table),
    ]  # type: List[Stripped]
    errors = []  # type: List[Error]

    for our_type in symbol_table.our_types:
        if isinstance(our_type, intermediate.Enumeration):
            enum_blocks = _generate_to_pb_for_enum(
                enum=our_type, aas_module=aas_module, aas_pb_module=aas_pb_module
            )
            blocks.extend(enum_blocks)
        elif isinstance(our_type, intermediate.ConstrainedPrimitive):
            # NOTE (mristin):
            # We do not represent constrained primitives in Python types.
            pass
        elif isinstance(
            our_type, (intermediate.AbstractClass, intermediate.ConcreteClass)
        ):
            if len(our_type.concrete_descendants) > 0:
                blocks.extend(
                    _generate_to_pb_choice(
                        cls=our_type, aas_module=aas_module, aas_pb_module=aas_pb_module
                    )
                )

            if isinstance(our_type, intermediate.ConcreteClass):
                blocks.append(
                    _generate_concrete_to_pb_for_class(
                        cls=our_type, aas_module=aas_module, aas_pb_module=aas_pb_module
                    )
                )
        else:
            assert_never(our_type)

    blocks.extend(
        _generate_general_to_pb(
            symbol_table=symbol_table,
            aas_module=aas_module,
            aas_pb_module=aas_pb_module,
        )
    )

    if len(errors) > 0:
        return None, errors

    return blocks, None


# endregion

_THIS_PATH = pathlib.Path(os.path.realpath(__file__))


@ensure(lambda result: not (result[0] is not None) or (result[0].endswith("\n")))
@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
def _generate(
    symbol_table: intermediate.SymbolTable,
    spec_impls: specific_implementations.SpecificImplementations,
) -> Tuple[Optional[str], Optional[List[Error]]]:
    """
    Generate the content of the pbization module.

    The ``aas_module`` indicates the fully-qualified name of the base SDK module.

    The ``aas_pb_module`` indicates the fully-qualified name of this base module.
    """
    aas_module_key = specific_implementations.ImplementationKey(
        "base_qualified_module_name.txt"
    )

    aas_module_text = spec_impls.get(aas_module_key, None)
    if aas_module_text is None:
        return None, [
            Error(
                None,
                f"The implementation snippet for the base qualified module name "
                f"is missing: {aas_module_key}",
            )
        ]

    if not python_common.QUALIFIED_MODULE_NAME_RE.fullmatch(aas_module_text):
        return None, [
            Error(
                None,
                f"The text from the snippet {aas_module_key} "
                f"is not a valid qualified module name: {aas_module_text!r}\n",
            )
        ]

    aas_module = python_common.QualifiedModuleName(aas_module_text)

    aas_pb_module_key = specific_implementations.ImplementationKey(
        "qualified_module_name_for_protobuf_library.txt"
    )

    aas_pb_module_text = spec_impls.get(aas_pb_module_key, None)
    if aas_pb_module_text is None:
        return None, [
            Error(
                None,
                f"The implementation snippet for the qualified name of the protobuf "
                f"library is missing: {aas_pb_module_key}",
            )
        ]

    if not python_common.QUALIFIED_MODULE_NAME_RE.fullmatch(aas_pb_module_text):
        return None, [
            Error(
                None,
                f"The text from the snippet {aas_pb_module_key} "
                f"is not a valid qualified module name: {aas_pb_module_text!r}\n",
            )
        ]

    aas_pb_module = python_common.QualifiedModuleName(aas_pb_module_text)

    warning = Stripped(
        f"""\
# Automatically generated with {_THIS_PATH.parent.name}/{_THIS_PATH.name}.
# Do NOT edit or append."""
    )

    blocks = [
        warning,
        Stripped('"""Convert instances from and to Protocol Buffers."""'),
        Stripped(
            f"""\
from typing import Mapping, TypeVar

import google.protobuf.message

from {aas_module} import types
import {aas_pb_module}.types_pb2 as types_pb""",
        ),
        Stripped("# region From Protocol Buffers"),
    ]  # type: List[Stripped]

    errors = []  # type: List[Error]

    from_pb_blocks, from_pb_errors = _generate_from_pb(
        symbol_table=symbol_table, aas_pb_module=aas_pb_module
    )
    if from_pb_errors is not None:
        errors.append(
            Error(
                None,
                "Failed to generate one or more from-protobuf conversion",
                underlying=errors,
            )
        )
    else:
        assert from_pb_blocks is not None
        blocks.extend(from_pb_blocks)

    blocks.extend(
        [
            Stripped("# endregion From Protocol Buffers"),
            Stripped("# region To Protocol Buffers"),
        ]
    )

    to_pb_blocks, to_pb_errors = _generate_to_pb(
        symbol_table=symbol_table, aas_module=aas_module, aas_pb_module=aas_pb_module
    )
    if to_pb_errors is not None:
        errors.append(
            Error(
                None,
                "Failed to generate one or more to-protobuf conversion",
                underlying=errors,
            )
        )
    else:
        assert to_pb_blocks is not None
        blocks.extend(to_pb_blocks)

    blocks.append(
        Stripped("# endregion To Protocol Buffers"),
    )

    blocks.append(warning)

    out = io.StringIO()
    for i, block in enumerate(blocks):
        if i > 0:
            out.write("\n\n\n")

        out.write(block)

    out.write("\n")

    return out.getvalue(), None


def execute(
    context: run.Context,
    stdout: TextIO,
    stderr: TextIO,
) -> int:
    """
    Execute the generation with the given parameters.

    Return the error code, or 0 if no errors.
    """
    code, errors = _generate(
        symbol_table=context.symbol_table, spec_impls=context.spec_impls
    )
    if errors is not None:
        run.write_error_report(
            message=f"Failed to generate the Python-Protobuf library "
            f"based on {context.model_path}",
            errors=[context.lineno_columner.error_message(error) for error in errors],
            stderr=stderr,
        )
        return 1

    assert code is not None

    pth = context.output_dir / "pbization.py"
    try:
        pth.write_text(code, encoding="utf-8")
    except Exception as exception:
        run.write_error_report(
            message=f"Failed to write the Python-Protobuf library to {pth}",
            errors=[str(exception)],
            stderr=stderr,
        )
        return 1

    stdout.write(f"Code generated to: {context.output_dir}\n")

    return 0



"""Generate TypeScript constants corresponding to the constants of the meta-model."""
import io
from typing import (
    Optional,
    List,
    Tuple,
)

from icontract import ensure

from aas_core_codegen import intermediate
from aas_core_codegen.common import (
    Error,
    assert_never,
    Stripped,
    indent_but_first_line,
)
from aas_core_codegen.typescript import (
    common as typescript_common,
    naming as typescript_naming,
    description as typescript_description,
)
from aas_core_codegen.typescript.common import (
    INDENT as I,
)


# region Generation


def _generate_documentation_comment_for_constant(
    description: intermediate.DescriptionOfConstant,
    context: typescript_description.Context,
) -> Tuple[Optional[Stripped], Optional[List[Error]]]:
    """Generate the documentation comment for the given constant."""
    # fmt: off
    comment, errors = (
        typescript_description
        .generate_documentation_comment_for_summary_remarks(
            description=description, context=context
        )
    )
    # fmt: on

    if errors is not None:
        return None, errors

    assert comment is not None
    return comment, None


@ensure(lambda result: (result[0] is None) ^ (result[1] is None))
def _generate_constant_primitive(
    constant: intermediate.ConstantPrimitive,
) -> Tuple[Optional[Stripped], Optional[Error]]:
    """Generate the definition of a constant primitive."""
    writer = io.StringIO()

    if constant.description is not None:
        comment, comment_errors = _generate_documentation_comment_for_constant(
            description=constant.description,
            context=typescript_description.Context(
                module=typescript_common.CONSTANTS_MODULE, cls_or_enum=None
            ),
        )
        if comment_errors is not None:
            return None, Error(
                constant.parsed.node,
                f"Failed to generate the documentation comment for {constant.name!r}",
                comment_errors,
            )

        assert comment is not None
        writer.write(comment)
        writer.write("\n")

    constant_name = typescript_naming.constant_name(constant.name)

    if constant.a_type is intermediate.PrimitiveType.BOOL:
        assert isinstance(constant.value, bool)
        literal = typescript_common.boolean_literal(constant.value)
        writer.write(f"export const {constant_name} = {literal};")

    elif constant.a_type is intermediate.PrimitiveType.INT:
        assert isinstance(constant.value, int)

        if not typescript_common.representable_as_number(constant.value):
            return None, Error(
                constant.parsed.node,
                f"The integer can not be represented as a double-precision "
                f"floating-point number in TypeScript: {constant.value!r}",
            )

        literal = typescript_common.numeric_literal(constant.value)

        writer.write(f"export const {constant_name} = {literal};")

    elif constant.a_type is intermediate.PrimitiveType.FLOAT:
        assert isinstance(constant.value, float)

        # NOTE (mristin, 2022-11-11):
        # We assume that the float constants are not meant to be all to precise.
        # Therefore, we use a string representation here. However, beware that we
        # might have to use a more precise representation in the future if the spec
        # change.
        literal = typescript_common.numeric_literal(constant.value)

        writer.write(f"export const {constant_name} = {literal};")

    elif constant.a_type is intermediate.PrimitiveType.STR:
        assert isinstance(constant.value, str)
        literal = typescript_common.string_literal(constant.value)

        writer.write(f"export const {constant_name} = {literal};")

    elif constant.a_type is intermediate.PrimitiveType.BYTEARRAY:
        assert isinstance(constant.value, bytearray)

        literal, _ = typescript_common.bytes_literal(value=constant.value)

        writer.write(
            f"""\
export const {constant_name} =
{I}{indent_but_first_line(literal, I)};"""
        )

    else:
        assert_never(constant.a_type)

    return Stripped(writer.getvalue()), None


@ensure(lambda result: (result[0] is None) ^ (result[1] is None))
def _generate_constant_set_of_primitives(
    constant: intermediate.ConstantSetOfPrimitives,
) -> Tuple[Optional[Stripped], Optional[Error]]:
    """Generate the definition of a constant set of primitives."""
    errors = []  # type: List[Error]

    writer = io.StringIO()

    if constant.description is not None:
        comment, comment_errors = _generate_documentation_comment_for_constant(
            description=constant.description,
            context=typescript_description.Context(
                module=typescript_common.CONSTANTS_MODULE, cls_or_enum=None
            ),
        )
        if comment_errors is not None:
            errors.append(
                Error(
                    constant.parsed.node,
                    f"Failed to generate the documentation comment for {constant.name!r}",
                    comment_errors,
                )
            )
        else:
            assert comment is not None
            writer.write(comment)
            writer.write("\n")

    literal_codes = []  # type: List[str]
    set_type: str

    if constant.a_type is intermediate.PrimitiveType.BOOL:
        set_type = "boolean"
        for literal in constant.literals:
            assert isinstance(literal.value, bool)
            literal_codes.append(typescript_common.boolean_literal(literal.value))

    elif constant.a_type is intermediate.PrimitiveType.INT:
        set_type = "number"
        for literal in constant.literals:
            assert isinstance(literal.value, int)

            if not typescript_common.representable_as_number(literal.value):
                errors.append(
                    Error(
                        literal.parsed.node,
                        f"The item of the constant set {constant.name!r} can not be represented as "
                        f"a double-precision floating-point number: {literal.value!r}",
                    )
                )
            else:
                literal_codes.append(typescript_common.numeric_literal(literal.value))

    elif constant.a_type is intermediate.PrimitiveType.FLOAT:
        set_type = "number"

        for literal in constant.literals:
            assert isinstance(literal.value, float)

            # NOTE (mristin, 2022-11-12):
            # We assume that the float constants are not meant to be all to precise.
            # Therefore, we use a string representation here. However, beware that we
            # might have to use a more precise representation in the future if the spec
            # change.
            literal_codes.append(typescript_common.numeric_literal(literal.value))

    elif constant.a_type is intermediate.PrimitiveType.STR:
        set_type = "string"

        for literal in constant.literals:
            assert isinstance(literal.value, str)
            literal_codes.append(typescript_common.string_literal(literal.value))

    elif constant.a_type is intermediate.PrimitiveType.BYTEARRAY:
        errors.append(
            Error(
                constant.parsed.node,
                f"TypeScript does not support sets of byte arrays, so we can not transpile "
                f"the constant set {constant.name!r}",
            )
        )

    else:
        assert_never(constant.a_type)

    literals_joined = ",\n".join(literal_codes)

    constant_name = typescript_naming.constant_name(constant.name)

    writer.write(
        f"""\
export const {constant_name} = new Set<{set_type}>([
{I}{indent_but_first_line(literals_joined, I)}
]);"""
    )

    return Stripped(writer.getvalue()), None


@ensure(lambda result: (result[0] is None) ^ (result[1] is None))
def _generate_constant_set_of_enumeration_literals(
    constant: intermediate.ConstantSetOfEnumerationLiterals,
) -> Tuple[Optional[Stripped], Optional[Error]]:
    """Generate the definition of a constant set of enumeration literals."""
    writer = io.StringIO()

    if constant.description is not None:
        comment, comment_errors = _generate_documentation_comment_for_constant(
            description=constant.description,
            context=typescript_description.Context(
                module=typescript_common.CONSTANTS_MODULE, cls_or_enum=None
            ),
        )
        if comment_errors is not None:
            return None, Error(
                constant.parsed.node,
                f"Failed to generate the documentation comment for {constant.name!r}",
                comment_errors,
            )

        assert comment is not None
        writer.write(comment)
        writer.write("\n")

    enum_name = typescript_naming.enum_name(constant.enumeration.name)

    literal_codes = []  # type: List[str]

    for literal in constant.literals:
        literal_name = typescript_naming.enum_literal_name(literal.name)

        literal_codes.append(f"AasTypes.{enum_name}.{literal_name}")

    constant_name = typescript_naming.constant_name(constant.name)

    literals_joined = ",\n".join(literal_codes)

    writer.write(
        f"""\
export const {constant_name} = new Set<AasTypes.{enum_name}>([
{I}{indent_but_first_line(literals_joined, I)}
]);"""
    )

    return Stripped(writer.getvalue()), None


# fmt: off
@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
@ensure(
    lambda result:
    not (result[0] is not None) or result[0].endswith('\n'),
    "Trailing newline mandatory for valid end-of-files"
)
# fmt: on
def generate(
    symbol_table: intermediate.SymbolTable,
) -> Tuple[Optional[str], Optional[List[Error]]]:
    """Generate the TypeScript code of the constants based on the symbol table."""
    errors = []  # type: List[Error]

    blocks = [
        Stripped(
            """\
/**
 * Provide constant values of the meta-model.
 */"""
        ),
        typescript_common.WARNING,
        Stripped("/* eslint-disable @typescript-eslint/no-unused-vars */"),
        Stripped('import * as AasTypes from "./types";'),
    ]  # type: List[Stripped]

    for constant in symbol_table.constants:
        block: Optional[Stripped]
        error: Optional[Error]

        if isinstance(constant, intermediate.ConstantPrimitive):
            block, error = _generate_constant_primitive(constant=constant)
        elif isinstance(constant, intermediate.ConstantSetOfPrimitives):
            block, error = _generate_constant_set_of_primitives(constant=constant)
        elif isinstance(constant, intermediate.ConstantSetOfEnumerationLiterals):
            block, error = _generate_constant_set_of_enumeration_literals(
                constant=constant
            )
        else:
            assert_never(constant)

        if error is not None:
            errors.append(error)
            continue

        assert block is not None
        blocks.append(block)

    if len(errors) > 0:
        return None, errors

    blocks.append(Stripped("/* eslint-enable @typescript-eslint/no-unused-vars */"))
    blocks.append(typescript_common.WARNING)

    writer = io.StringIO()
    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")

        writer.write(block)

    writer.write("\n")

    return writer.getvalue(), None


# endregion



"""Provide common functions shared among different C++ code generation modules."""

import io
import math
import re
from typing import List, Tuple, Optional

from icontract import ensure, require

from aas_core_codegen import intermediate
from aas_core_codegen.common import (
    Stripped,
    assert_never,
    Identifier,
    indent_but_first_line,
)
from aas_core_codegen.cpp import (
    naming as cpp_naming,
)

# See: https://google.github.io/styleguide/cppguide.html#Spaces_vs._Tabs
INDENT = "  "
INDENT2 = INDENT * 2
INDENT3 = INDENT * 3
INDENT4 = INDENT * 4
INDENT5 = INDENT * 5
INDENT6 = INDENT * 6
INDENT7 = INDENT * 7
INDENT8 = INDENT * 8

WARNING = Stripped(
    """\
// This code has been automatically generated by aas-core-codegen.
// Do NOT edit or append."""
)


@ensure(lambda result: result.startswith('L"'))
@ensure(lambda result: result.endswith('"'))
def wstring_literal(text: str) -> Stripped:
    """Generate a C++ string literal from the ``text``."""
    escaped = []  # type: List[str]

    for character in text:
        code_point = ord(character)

        if character == "\a":
            escaped.append("\\a")
        elif character == "\b":
            escaped.append("\\b")
        elif character == "\f":
            escaped.append("\\f")
        elif character == "\n":
            escaped.append("\\n")
        elif character == "\r":
            escaped.append("\\r")
        elif character == "\t":
            escaped.append("\\t")
        elif character == "\v":
            escaped.append("\\v")
        elif character == '"':
            escaped.append('\\"')
        elif character == "\\":
            escaped.append("\\\\")
        elif code_point < 32:
            # Non-printable ASCII characters
            escaped.append(f"\\x{ord(character):x}")
        elif code_point <= 127:
            # ASCII
            escaped.append(character)
        elif 127 < code_point < 255:
            # Above ASCII, but can be encoded as a single byte
            escaped.append(f"\\x{ord(character):x}")
        elif 255 <= code_point < 65536:
            # Above ASCII
            escaped.append(f"\\u{ord(character):04x}")
        elif code_point >= 65536:
            # Above Unicode Binary Multilingual Pane
            escaped.append(f"\\U{ord(character):08x}")
        else:
            escaped.append(character)

    # NOTE (mristin, 2023-06-28):
    # We use std::wstring, therefore ``L`` prefix.
    return Stripped('L"{}"'.format("".join(escaped)))


@require(lambda character: len(character) == 1)
def wchar_literal(character: str) -> Stripped:
    """Generate a C++ wide character literal from the ``character``."""
    code_point = ord(character)

    escaped: str

    if character == "\a":
        escaped = "L'\\a'"
    elif character == "\b":
        escaped = "L'\\b'"
    elif character == "\f":
        escaped = "L'\\f'"
    elif character == "\n":
        escaped = "L'\\n'"
    elif character == "\r":
        escaped = "L'\\r'"
    elif character == "\t":
        escaped = "L'\\t'"
    elif character == "\v":
        escaped = "L'\\v'"
    elif character == "'":
        escaped = "L'\\''"
    elif character == "\\":
        escaped = "L'\\\\'"
    elif code_point < 32:
        # Non-printable ASCII characters
        escaped = f"L'\\x{ord(character):x}'"
    elif code_point <= 127:
        # ASCII
        escaped = f"L'{character}'"
    elif 127 < code_point < 255:
        # Above ASCII, but can be encoded as a single byte
        escaped = f"L'\\x{ord(character):x}'"
    elif 0xD800 <= code_point <= 0xDFFF:
        # NOTE (mristin):
        # These are the surrogate points and can not be represented as wide character
        # literals directly as common compilers such as MSVC++ will complain.
        #
        # We have to fool the compiler at this point as we deliberately want to model
        # the surrogate point.
        escaped = f"static_cast<wchar_t>(0x{ord(character):04x})"

    # NOTE (mristin):
    # Mind the intersecting range for surrogate points just above if you ever convert
    # this if-elif-else statement into a mapping or pattern matching.
    elif 255 <= code_point < 65536:
        # Above ASCII
        escaped = f"L'\\u{ord(character):04x}'"
    elif code_point >= 65536:
        # Above Unicode Basic Multilingual Pane
        escaped = f"L'\\U{ord(character):08x}'"
    else:
        raise AssertionError(f"Unexpected unhandled character: {character!r}")

    return Stripped(escaped)


# fmt: off
# NOTE (mristin, 2023-09-24):
# We use a pre-condition here to simplify the client code. The client must check
# before if the input text is all in ASCII, and report to the user if there are
# any non-ASCII characters in the input.
@require(
    lambda text:
    all(
        ord(character) <= 127
        for character in text
    ),
    "Only ASCII text can be converted to a C++ string literal, otherwise encoding "
    "must be assumed."
)
@ensure(lambda result: result.startswith('"'))
@ensure(lambda result: result.endswith('"'))
# fmt: on
def string_literal(text: str) -> Stripped:
    """Generate a C++ string literal from the ``text``."""
    escaped = []  # type: List[str]

    for character in text:
        code_point = ord(character)

        if character == "\a":
            escaped.append("\\a")
        elif character == "\b":
            escaped.append("\\b")
        elif character == "\f":
            escaped.append("\\f")
        elif character == "\n":
            escaped.append("\\n")
        elif character == "\r":
            escaped.append("\\r")
        elif character == "\t":
            escaped.append("\\t")
        elif character == "\v":
            escaped.append("\\v")
        elif character == '"':
            escaped.append('\\"')
        elif character == "\\":
            escaped.append("\\\\")
        elif code_point < 32:
            # Non-printable ASCII characters
            escaped.append(f"\\x{ord(character):x}")
        elif code_point <= 127:
            escaped.append(character)
        else:
            # Above ASCII
            raise ValueError(r"Unexpected non-ascii code point: {character!r}")

    return Stripped('"{}"'.format("".join(escaped)))


def needs_escaping(text: str) -> bool:
    """Check whether the ``text`` contains a character that needs escaping."""
    for character in text:
        code_point = ord(character)

        if character == "\a":
            return True
        elif character == "\b":
            return True
        elif character == "\f":
            return True
        elif character == "\n":
            return True
        elif character == "\r":
            return True
        elif character == "\t":
            return True
        elif character == "\v":
            return True
        elif character == '"':
            return True
        elif character == "\\":
            return True
        elif code_point < 32:
            # Non-printable ASCII characters
            return True
        elif code_point <= 127:
            # ASCII
            continue
        elif 127 < code_point < 255:
            # Above ASCII, but can be encoded as a single byte
            return True
        elif 255 <= code_point < 65536:
            # Above ASCII
            return True
        elif code_point >= 65536:
            # Above Unicode Binary Multilingual Pane
            return True
        else:
            raise AssertionError(f"Unexpected unhandled character: {character!r}")

    return False


def boolean_literal(value: bool) -> Stripped:
    """Generate the boolean literal corresponding to the ``value``."""
    return Stripped("true") if value else Stripped("false")


def float_literal(value: float) -> Stripped:
    """Generate the float literal.

    We assume that the precision of the literal is not critical and rely on
    Python's ``str(.)`` function. However, if you want to specify the exact
    number, you have to format the number yourself, probably using G17 representation.
    """
    if math.isnan(value):
        return Stripped("std::numeric_limits<double>::quiet_NaN()")
    if value == math.inf:
        return Stripped("std::numeric_limits<double>::infinity()")
    elif value == -math.inf:
        return Stripped("-std::numeric_limits<double>::infinity()")
    else:
        return Stripped(str(value))


def bytes_literal(value: bytes) -> Tuple[Stripped, bool]:
    """
    Generate an expression representing the ``value``.

    If there are more than 8 bytes, a multi-line expression is returned.

    :param value: to be represented
    :return: (C++ expression, is multi-line)
    """
    if len(value) == 0:
        return Stripped("std::vector<std::uint8_t>()"), False

    writer = io.StringIO()

    # noinspection PyUnusedLocal
    multi_line = None  # type: Optional[bool]

    if len(value) <= 8:
        items_joined = ", ".join(f"0x{byte:02x}" for byte in value)
        return Stripped(f"{{{items_joined}}}"), False
    else:
        writer.write(
            """\
{"""
        )

        for start in range(0, len(value), 8):
            if start == 0:
                writer.write(f"\n{INDENT}")
            else:
                writer.write(f",\n{INDENT}")

            end = min(start + 8, len(value))

            assert start < end

            for i, byte in enumerate(value[start:end]):
                if i > 0:
                    writer.write(", ")

                writer.write(f"0x{byte:02x}")

        writer.write("\n}")
        multi_line = True

        return Stripped(writer.getvalue()), multi_line


PRIMITIVE_TYPE_MAP = {
    intermediate.PrimitiveType.BOOL: Stripped("bool"),
    intermediate.PrimitiveType.INT: Stripped("int64_t"),
    intermediate.PrimitiveType.FLOAT: Stripped("double"),
    intermediate.PrimitiveType.STR: Stripped("std::wstring"),
    intermediate.PrimitiveType.BYTEARRAY: Stripped("std::vector<std::uint8_t>"),
}


def _assert_all_primitive_types_are_mapped() -> None:
    """Assert that we have explicitly mapped all the primitive types to C++."""
    all_primitive_literals = set(literal.value for literal in PRIMITIVE_TYPE_MAP)

    mapped_primitive_literals = set(
        literal.value for literal in intermediate.PrimitiveType
    )

    all_diff = all_primitive_literals.difference(mapped_primitive_literals)
    mapped_diff = mapped_primitive_literals.difference(all_primitive_literals)

    messages = []  # type: List[str]
    if len(mapped_diff) > 0:
        messages.append(
            f"More primitive maps are mapped than there were defined "
            f"in the ``intermediate._types``: {sorted(mapped_diff)}"
        )

    if len(all_diff) > 0:
        messages.append(
            f"One or more primitive types in the ``intermediate._types`` were not "
            f"mapped in PRIMITIVE_TYPE_MAP: {sorted(all_diff)}"
        )

    if len(messages) > 0:
        raise AssertionError("\n\n".join(messages))


_assert_all_primitive_types_are_mapped()

COMMON_NAMESPACE = Identifier("common")
TYPES_NAMESPACE = Identifier("types")
CONSTANTS_NAMESPACE = Identifier("constants")
VERIFICATION_NAMESPACE = Identifier("verification")
JSONIZATION_NAMESPACE = Identifier("jsonization")
XMLIZATION_NAMESPACE = Identifier("xmlization")
REVM_NAMESPACE = Identifier("revm")
PATTERN_NAMESPACE = Identifier("pattern")


def generate_primitive_type(primitive_type: intermediate.PrimitiveType) -> Stripped:
    """Generate the C++ type for the given primitive type."""
    return PRIMITIVE_TYPE_MAP[primitive_type]


def primitive_type_is_referencable(primitive_type: intermediate.PrimitiveType) -> bool:
    """Return ``True`` if the primitive type denotes a referencable value in C++."""
    if primitive_type is intermediate.PrimitiveType.BOOL:
        return False

    elif primitive_type is intermediate.PrimitiveType.INT:
        return False

    elif primitive_type is intermediate.PrimitiveType.FLOAT:
        return False

    elif primitive_type is intermediate.PrimitiveType.STR:
        return True

    elif primitive_type is intermediate.PrimitiveType.BYTEARRAY:
        return True
    else:
        assert_never(primitive_type)


def generate_primitive_type_with_const_ref_if_applicable(
    primitive_type: intermediate.PrimitiveType,
) -> Stripped:
    """Generate the C++ type and wrap it in ``const T&``, if applicable."""
    code = generate_primitive_type(primitive_type)

    if primitive_type_is_referencable(primitive_type):
        return Stripped(f"const {code}&")

    return code


def generate_type(
    type_annotation: intermediate.TypeAnnotationUnion,
    types_namespace: Optional[Identifier] = None,
    common_namespace: Optional[Identifier] = COMMON_NAMESPACE,
) -> Stripped:
    """
    Generate the C++ type for the given type annotation.

    If ``types_namespace`` is specified, it is prepended to all our types.

    If `common_namespace` is specified, it is prepended to ``optional``.
    """
    if isinstance(type_annotation, intermediate.PrimitiveTypeAnnotation):
        return generate_primitive_type(type_annotation.a_type)

    elif isinstance(type_annotation, intermediate.OurTypeAnnotation):
        our_type = type_annotation.our_type

        if isinstance(our_type, intermediate.Enumeration):
            enum_name = cpp_naming.enum_name(type_annotation.our_type.name)
            if types_namespace is None:
                return enum_name

            return Stripped(f"{types_namespace}::{enum_name}")

        elif isinstance(our_type, intermediate.ConstrainedPrimitive):
            return generate_primitive_type(our_type.constrainee)

        elif isinstance(
            our_type, (intermediate.AbstractClass, intermediate.ConcreteClass)
        ):
            # NOTE (mristin, 2023-06-28):
            # We always refer to interfaces even in cases of concrete classes without
            # concrete descendants since we want to allow enhancing.
            interface_name = cpp_naming.interface_name(our_type.name)

            type_identifier: str

            if types_namespace is None:
                type_identifier = interface_name
            else:
                type_identifier = f"{types_namespace}::{interface_name}"

            assert not type_identifier.endswith(">")
            return Stripped(f"std::shared_ptr<{type_identifier}>")

    elif isinstance(type_annotation, intermediate.ListTypeAnnotation):
        item_type = generate_type(
            type_annotation=type_annotation.items, types_namespace=types_namespace
        )

        if "<" not in item_type:
            return Stripped(f"std::vector<{item_type}>")

        return Stripped(
            f"""\
std::vector<
{INDENT}{indent_but_first_line(item_type, INDENT)}
>"""
        )

    elif isinstance(type_annotation, intermediate.OptionalTypeAnnotation):
        value_type = generate_type(
            type_annotation=type_annotation.value, types_namespace=types_namespace
        )

        common_namespace_prefix = (
            "" if common_namespace is None else f"{common_namespace}::"
        )

        if "<" not in value_type:
            return Stripped(f"{common_namespace_prefix}optional<{value_type}>")

        return Stripped(
            f"""\
{common_namespace_prefix}optional<
{INDENT}{indent_but_first_line(value_type, INDENT)}
>"""
        )

    else:
        assert_never(type_annotation)

    raise AssertionError("Should not have gotten here")


def is_referencable(type_annotation: intermediate.TypeAnnotationUnion) -> bool:
    """Return ``True`` if the type annotation denotes a referencable value in C++."""
    if isinstance(type_annotation, intermediate.OptionalTypeAnnotation):
        return True

    primitive_type = intermediate.try_primitive_type(type_annotation)
    if primitive_type is not None:
        return primitive_type_is_referencable(primitive_type)

    if isinstance(type_annotation, intermediate.OurTypeAnnotation) and isinstance(
        type_annotation.our_type, intermediate.Enumeration
    ):
        return False

    return True


def generate_type_with_const_ref_if_applicable(
    type_annotation: intermediate.TypeAnnotationUnion,
    types_namespace: Optional[Identifier] = None,
) -> Stripped:
    """
    Generate the C++ type and wrap it in ``const T&``, if applicable.

    If ``types_namespace`` is specified, it is prepended to all our types.
    """
    code = generate_type(
        type_annotation=type_annotation, types_namespace=types_namespace
    )
    if is_referencable(type_annotation):
        return Stripped(f"const {code}&")

    return code


def generate_type_with_ref(
    type_annotation: intermediate.TypeAnnotationUnion,
    types_namespace: Optional[Identifier] = None,
) -> Stripped:
    """
    Generate the C++ type and wrap it as a reference in ``T&``.

    If ``types_namespace`` is specified, it is prepended to all our types.
    """
    code = generate_type(
        type_annotation=type_annotation, types_namespace=types_namespace
    )
    return Stripped(f"{code}&")


_ANGLE_BRACKETS_IN_TYPE_RE = re.compile(r"\s*([<>])\s*")
_INDENT_LIST = [
    INDENT,
    INDENT2,
    INDENT3,
    INDENT4,
    INDENT5,
    INDENT6,
    INDENT7,
    INDENT8,
]


def break_type_in_lines(text: str) -> str:
    """Break the given C++ type in multiple lines, if applicable."""
    if len(text) == 0:
        return ""

    buffer = io.StringIO()

    level = 0
    last_match = None  # type: Optional[re.Match[str]]
    for match in _ANGLE_BRACKETS_IN_TYPE_RE.finditer(text):
        if last_match is None:
            buffer.write(text[: match.start()])
        else:
            buffer.write(text[last_match.end() : match.start()])

        assert (
            0 <= level < len(_INDENT_LIST)
        ), f"Expected at most {len(_INDENT_LIST)} levels, but got: {level=}"

        angle_bracket = match.group(1)
        if angle_bracket == "<":
            buffer.write("<\n" + _INDENT_LIST[level])
            level += 1
        elif angle_bracket == ">":
            level -= 1
            buffer.write(" >")
        else:
            raise AssertionError(f"Unhandled angle bracket: {angle_bracket=}")

        last_match = match

    if last_match is None:
        # NOTE (mristin, 2024-01-18):
        # Not a signle angle bracket has been detected.
        return text

    buffer.write(text[last_match.end() :])

    return buffer.getvalue()


class GeneratorForLoopVariables:
    """
    Generate a unique variable name based on ``item`` stem.

    >>> generator = GeneratorForLoopVariables()

    >>> next(generator)
    'v'

    >>> next(generator)
    'v1'

    >>> next(generator)
    'v2'
    """

    def __init__(self) -> None:
        """Initialize with the zero counter."""
        self.counter = 0

    def __next__(self) -> Identifier:
        """Generate the next variable name."""
        if self.counter == 0:
            result = Identifier("v")
        else:
            result = Identifier(f"v{self.counter}")

        self.counter += 1

        return result


def generate_include_prefix_path(library_namespace: Stripped) -> Stripped:
    """
    Generate the prefix path for the includes.

    >>> generate_include_prefix_path(Stripped("some::name::space"))
    'some/name/space'
    """
    return Stripped(library_namespace.replace("::", "/"))


def generate_namespace_opening(library_namespace: Stripped) -> Stripped:
    """
    Generate the code to open the nested library namespace.

    >>> generate_namespace_opening(Stripped("some::name::space"))
    'namespace some {\\nnamespace name {\\nnamespace space {'
    """
    namespace_parts = library_namespace.split("::")
    return Stripped("\n".join(f"namespace {part} {{" for part in namespace_parts))


def generate_namespace_closing(library_namespace: Stripped) -> Stripped:
    """
    Generate the code to close the nested library namespace.

    >>> generate_namespace_closing(Stripped("some::name::space"))
    '}  // namespace space\\n}  // namespace name\\n}  // namespace some'
    """
    namespace_parts = library_namespace.split("::")
    return Stripped(
        "\n".join(f"}}  // namespace {part}" for part in reversed(namespace_parts))
    )


def include_guard_var(namespace: Stripped) -> Stripped:
    """
    Generate the variable name of the include guard.

    >>> include_guard_var(Stripped("aascore3::types"))
    'AASCORE3_TYPES_GUARD_'
    """
    return Stripped("_".join(namespace.split("::")).upper() + "_GUARD_")


def non_documentation_comment(text: str) -> str:
    r"""
    Generate the non-documentation comment prefixed with ``//``.

    >>> non_documentation_comment('test\nme')
    '// test\n// me'
    """
    lines = []  # type: List[str]
    for line in text.splitlines():
        stripped = line.strip()
        if len(stripped) > 0:
            lines.append(f"// {line}")
        else:
            lines.append("//")

    return "\n".join(lines)
